

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StrideRep': Generalises the idea of an array dimension       //
//                    stride.                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class StrideRep {
friend class Stride ;
friend class Range ;
public :
  StrideRep(const int _memStep) : memStep(_memStep) {}

  virtual ~StrideRep() {}

  virtual int offset(const Subscript*) = 0 ;
  virtual int step(Block&) = 0 ;

  virtual int volume() = 0 ;
protected :
  int memStep ;
private :
  int count ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `SimpleStrideRep'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class SimpleStrideRep : public StrideRep {
public :
  SimpleStrideRep(const int _memStep, const int _extent) :
        StrideRep(_memStep), extent(_extent) {}

  int offset(const Subscript*) ;
  int step(Block&) ;

  int volume() ;

protected :
  int extent ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `MultiblockStrideRep'                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class MultiblockStrideRep : public SimpleStrideRep {
public :
  MultiblockStrideRep(const int _memStep, const int _blkStep,
                      const int _extent) :
      SimpleStrideRep(_memStep, _extent), blkStep(_blkStep) {}

  int offset(const Subscript*) ;

protected :
  int blkStep ;

MEMORY_POOL_DECS
} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedStrideRep'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class PackedStrideRep : public SimpleStrideRep {
public :
  PackedStrideRep(const int _memStep, const int _stride, const int _extent) :
      SimpleStrideRep(_memStep, _extent), stride(_stride) {}

  int offset(const Subscript*) ;
  int step(Block&) ;

protected :
  int stride ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedMultiblockStrideRep'                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class PackedMultiblockStrideRep : public PackedStrideRep {
public :
  PackedMultiblockStrideRep(const int _memStep, const int _blkBase,
                            const int _blkStep, const int _stride,
                            const int _extent) :
    PackedStrideRep(_memStep, _stride, _extent),
    blkBase(_blkBase), blkStep(_blkStep) {}

  int offset(const Subscript*) ;

protected :
  int blkStep ;
  int blkBase ;

MEMORY_POOL_DECS
} ;



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PromoteStrideRep': Stride used to give a distributed         //
//                           ``view'' of a sequential data array.      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class PromoteStrideRep : public StrideRep {
public :
  PromoteStrideRep(RangeRep* x, const int oldStr) :
                       StrideRep(oldStr), parent(x) {
    parent->count++ ;
  }

  int offset(const Subscript*) ;
  int step(Block&) ;
  int volume() ;

  ~PromoteStrideRep() {
    if(parent && --parent->count == 0)
      delete parent ;
  }

private :
  RangeRep* parent ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ExtendedStrideRep'                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ExtendedStrideRep : public SimpleStrideRep {
public :
  ExtendedStrideRep(const int _memStep, const int _extent,
		    const int _ghostLo, const int _ghostHi) :
	SimpleStrideRep(_memStep, _extent + _ghostLo + _ghostHi),
	shift(_ghostLo) {}

  int offset(const Subscript*) ;

protected :
  int shift ;

MEMORY_POOL_DECS
} ;


// ---------------------------------------------------
//
// Things to do:
//
//     Modernize `PromoteStrideClass'.

