
// Operations on block cyclic distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockCyclicRangeRep': Block-cyclically distributed range.    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Constructor for primitive ranges

BlockCyclicRangeRep :: BlockCyclicRangeRep(int _extent,
                                           Dimension _dim, int _blockSize,
                                           Layout _layout) :
     RangeRep(_extent, _dim), blockSize(_blockSize), layout(_layout) {

  layerSize = dim.size() * blockSize ;
}


RangeRep* BlockCyclicRangeRep :: prm() {return this ; }
int BlockCyclicRangeRep :: bas() {return 0 ; }
int BlockCyclicRangeRep :: str() {return 1 ; }

int BlockCyclicRangeRep :: isScalar() {return 0 ; }


inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }

// Subscripting of Range.

Subscript* BlockCyclicRangeRep :: subscr(const int _i) {
  crd = (_i % layerSize) / blockSize ;
  blk = _i / layerSize ;
  sub = _i % blockSize ;

  return this ;
}

int BlockCyclicRangeRep :: subscr(const Subscript* s) {

  return layerSize * s->blk + blockSize * s->crd + s->sub ;
}


// Creation of child ranges and strides...

RangeRep* BlockCyclicRangeRep :: subrng(const int _extent,
                                        const int _base, const int _stride) {
  return new BlockCyclicSubRangeRep(_extent, _base, _stride, this,
                                    dim, blockSize, layout) ;
}

StrideRep* BlockCyclicRangeRep :: vector(const int dataLen) {
  const int blkTop   = (extent - 1) / layerSize ;
  const int blkStep = blockSize * dataLen ;

  return new MultiblockStrideRep(dataLen, blkStep, (blkTop + 1) * blkStep) ;
      // Allocates a layer even if extent = 0.
}


// Auxilliary functions for use during block iteration

void BlockCyclicRangeRep :: blockParams(Block& b, int i_lo, int i_hi) {
  b.blockCount   = i_hi - i_lo + 1 ;

  b.blockBase    = i_lo ;

  b.blockBaseSub = 0 ;
}

void BlockCyclicRangeRep :: blockInit(Block& b) {
  b.dim          = dim ;

  b.blockStep    = 1 ;

  b.blockStepSub = 1 ;
}


// Iteration over locally held blocks.

void BlockCyclicRangeRep :: locBlockInit(Block& b) {
  b.crd = dim.crd() ;
}

void BlockCyclicRangeRep :: locBlocksEnum(Block& b) {
  int i_lo, i_hi ;

  b.blk = 0 ;

  i_lo = b.crd * blockSize ;
  i_hi = min(i_lo + blockSize, extent) - 1 ;

  blockParams(b, i_lo, i_hi) ;
}

int BlockCyclicRangeRep :: locBlocksTest(Block& b) {
  return b.blockBase < extent ;
}

void BlockCyclicRangeRep :: locBlocksNext(Block& b) {
  int i_lo, i_hi ;

  b.blk++ ;

  i_lo = b.blockBase + layerSize ;
  i_hi = min(i_lo + blockSize, extent) - 1 ;

  blockParams(b, i_lo, i_hi) ;
}


// Iteration over all blocks of range.

void BlockCyclicRangeRep :: allBlocksEnum(Block& b) {
  int i_lo, i_hi ;

  b.blk = 0 ;
  b.crd = 0 ;

  i_lo = 0 ;
  i_hi = min(blockSize, extent) - 1 ;

  blockParams(b, i_lo, i_hi) ;
}

int BlockCyclicRangeRep :: allBlocksTest(Block& b) {
  return b.blockBase < extent ;
}

void BlockCyclicRangeRep :: allBlocksNext(Block& b) {
  int i_lo, i_hi ;

  b.crd++ ;
  if(b.crd == dim.size()) {
    b.crd = 0 ;
    b.blk++ ;
  }

  i_lo = b.blockBase + blockSize ;
  i_hi = min(i_lo + blockSize, extent) - 1 ;

  blockParams(b, i_lo, i_hi) ;
}


// Memory management

MEMORY_POOL_DEFN(BlockCyclicRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockCyclicSubRangeRep': The most general HPF range,         //
//                      including block-cyclic distribution, and       //
//                      arbitrary strided alignment.                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Constructor for derived ranges

BlockCyclicSubRangeRep :: BlockCyclicSubRangeRep(const int _extent,
			      const int _base, const int _stride,
                              BlockCyclicRangeRep* _parent,
                              Dimension _dim, int _blockSize, Layout _layout) :
				  BlockCyclicRangeRep(_extent,
                                                      _dim, _blockSize,
                                                      _layout),
				  base(_base), stride(_stride),
                                  parent(_parent) {
  parent->incr() ;
}

BlockCyclicSubRangeRep :: ~BlockCyclicSubRangeRep() {
  if(parent->decr() == 0) delete parent ;
}


RangeRep* BlockCyclicSubRangeRep :: prm() {return parent ; }
int BlockCyclicSubRangeRep :: bas() {return base ; }
int BlockCyclicSubRangeRep :: str() {return stride ; }


// Subscripting of Range.

Subscript* BlockCyclicSubRangeRep :: subscr(const int _i) {
  return BlockCyclicRangeRep :: subscr(base + stride * _i) ;
}

int BlockCyclicSubRangeRep :: subscr(const Subscript* s) {
  return (BlockCyclicRangeRep :: subscr(s) - base) / stride ;
}


// Creation of child ranges and strides...

RangeRep* BlockCyclicSubRangeRep :: subrng(const int _extent,
                                        const int _base, const int _stride) {
  return new BlockCyclicSubRangeRep(_extent,
                 base + stride * _base, stride * _stride, parent,
                 dim, blockSize, layout) ;
}

StrideRep* BlockCyclicSubRangeRep :: vector(const int dataLen) {
  if(layout == STANDARD) {
    const int blkTop  = (parent->size() - 1) / layerSize ;

    const int blkStep = blockSize * dataLen ;
    const int volume  = (blkTop + 1) * blockSize ;

    return new MultiblockStrideRep(dataLen, blkStep, volume) ;
	// Allocates a layer even if primitive extent == 0.
  }
  else {

    // ``Packed'' layouts.

    const int s_lo = base ;
    const int s_hi = base + (extent - 1) * stride ;

    if(stride == 0) AdlibError("BlockCyclicSubRangeRep :: vector: "
			       "range stride is zero") ;

    if(stride > 0) {
      const int blockWidth = (blockSize + stride - 1) / stride ;

      const int blkBase = (layout == ADAPT) ? 0 : s_lo / layerSize ;
      const int blkTop  = s_hi / layerSize ;

      const int blkStep = blockWidth * dataLen ;
      const int volume  = (blkTop - blkBase + 1) * blockWidth ;

      return new PackedMultiblockStrideRep(dataLen, blkBase, blkStep, stride,
                                           volume) ;
    }
    else {
      const int blockWidth = (blockSize - stride - 1) / (-stride) ;

      const int blkBase = (layout == ADAPT) ? 0 : s_hi / layerSize ;
      const int blkTop  = s_lo / layerSize ;

      const int blkStep = blockWidth * dataLen ;
      const int volume  = (blkTop - blkBase + 1) * blockWidth ;

      return new PackedMultiblockStrideRep(-dataLen, blkBase, blkStep, stride,
                                           volume) ;
    }

	  // Allocates a layer even if extent = 0.
  }
}


// Auxilliary functions for use during block iteration

void BlockCyclicSubRangeRep :: blockInit(Block& b) {
  b.dim          = dim ;

  b.blockStep    = 1 ;

  b.blockStepSub = stride ;
}

void BlockCyclicSubRangeRep :: blockParams(Block& b, int i_lo, int i_hi) {
  b.blockCount   = i_hi - i_lo + 1 ;

  b.blockBase    = i_lo ;

  b.blockBaseSub = (base + stride * i_lo) % blockSize ;
}

void BlockCyclicSubRangeRep :: blockRange(int& i_lo, int& i_hi,
                                          int _crd, int _blk) {

  // Define `i_lo' and `i_hi' for current values of `_blk', `_crd'

  // If `_blk' and `_crd' do not define a position within the bounds of
  // the range, the results are undefined but it is guaranteed that
  // i_hi < i_lo


  const int blockStart = _blk * layerSize + _crd * blockSize ;

  const int blockStartOffset  = blockStart - base ;
  const int blockFinishOffset = blockStartOffset + blockSize - 1 ;

  if(stride > 0) {
    i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
    if(blockFinishOffset >= 0)
      i_hi = min(blockFinishOffset / stride, extent - 1) ;
    else
      i_hi = -1 ;
  }
  else {
    i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
    if(blockStartOffset <= 0)
      i_hi = min(blockStartOffset / stride, extent - 1) ;
    else
      i_hi = -1 ;
  }
}


// Iteration over locally held blocks.

void BlockCyclicSubRangeRep :: locBlockInit(Block& b) {
  b.crd          = dim.crd() ;
}

void BlockCyclicSubRangeRep :: locBlocksEnum(Block& b) {
  int i_lo, i_hi ;

  b.blk = base / layerSize ;

  const int baseCoord = (base / blockSize) % dim.size() ;

  if(stride > 0 && b.crd < baseCoord) b.blk++ ;
  if(stride < 0 && b.crd > baseCoord) b.blk-- ;

  blockRange(i_lo, i_hi, b.crd, b.blk) ;

  if(i_lo > i_hi && i_lo < extent)
                 // ^^^^^^^^^^^^^^ makes assumptions about `i_lo'
                 //                when block is out of range
    locBlocksNext(b) ;
  else
    blockParams(b, i_lo, i_hi) ;
}

void BlockCyclicSubRangeRep :: locBlocksNext(Block& b) {
  int i_lo, i_hi ;

  do { 
    if(stride > 0)
      b.blk++ ;
    else
      b.blk-- ;

    blockRange(i_lo, i_hi, b.crd, b.blk) ;
  } while (i_lo > i_hi && i_lo < extent) ;
                       // ^^^^^^^^^^^^^^ makes assumptions about `i_lo'
                       //                when block is out of range

  blockParams(b, i_lo, i_hi) ;
}

// Iteration over all blocks of range.

void BlockCyclicSubRangeRep :: allBlocksEnum(Block& b) {
  int i_lo, i_hi ;

  b.blk = base / layerSize ;
  b.crd = (base / blockSize) % dim.size() ;

  blockRange(i_lo, i_hi, b.crd, b.blk) ;

  if(i_lo > i_hi && i_lo < extent)
                 // ^^^^^^^^^^^^^^ makes assumptions about `i_lo'
                 //                when block is out of range
    allBlocksNext(b) ;
  else
    blockParams(b, i_lo, i_hi) ;
}

void BlockCyclicSubRangeRep :: allBlocksNext(Block& b) {
  int i_lo, i_hi ;

  do {
    if(stride > 0) {
      b.crd++ ;
      if(b.crd == dim.size()) {
        b.crd = 0 ;
        b.blk++ ;
      }
    }
    else {
      b.crd-- ;
      if(b.crd == -1) {
        b.crd += dim.size() ;
        b.blk-- ;
      }
    }

    blockRange(i_lo, i_hi, b.crd, b.blk) ;

  } while (i_lo > i_hi && i_lo < extent) ;
                       // ^^^^^^^^^^^^^^ makes assumptions about `i_lo'
                       //                when block is out of range
  blockParams(b, i_lo, i_hi) ;
}


// Memory management

MEMORY_POOL_DEFN(BlockCyclicSubRangeRep)

// ---------------------------------------------------
//
// Things to do:
//
