
// Operations on block distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRangeRep': Block-distributed range.                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

BlockRangeRep :: BlockRangeRep(int _extent, Dimension _dim,
                               const int _blockSize) :
                     RangeRep(_extent, _dim), blockSize(_blockSize) {}

RangeRep* BlockRangeRep :: prm() {return this ; }
int BlockRangeRep :: bas() {return 0 ; }
int BlockRangeRep :: str() {return 1 ; }

int BlockRangeRep :: isScalar() {return 0 ; }


inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }


// Subscripting of Range.

Subscript* BlockRangeRep :: subscr(const int _i) {
  crd = _i / blockSize ;
  sub = _i % blockSize ;

  return this ;
}

int BlockRangeRep :: subscr(const Subscript* s) {
  return blockSize * s->crd + s->sub ;
}


// Creation of child ranges and strides...

RangeRep* BlockRangeRep :: subrng(const int _extent,
                                  const int _base, const int _stride) {
  return new BlockSubRangeRep(_extent, _base, _stride, this,
                              dim, blockSize) ;
}

StrideRep* BlockRangeRep :: vector(const int dataLen) {
  return new SimpleStrideRep(dataLen, blockSize) ;
}

StrideRep* BlockRangeRep :: subrngVector(const int dataLen, const int stride) {
  return vector(dataLen) ;
}


// Auxilliary functions for use during block iteration

void BlockRangeRep :: blockInit(Block& b) {
  b.dim          = dim ;

  b.blockStep    = 1 ;

  b.blockStepSub = 1 ;
}

void BlockRangeRep :: blockParams(Block& b, int i_lo, int i_hi) {
  b.blockCount   = i_hi - i_lo + 1 ;

  b.blockBase    = i_lo ;

  b.blockBaseSub = 0 ;
}


// Iteration over locally held blocks.

void BlockRangeRep :: locBlockInit(Block& b) {
  int i_lo, i_hi ;

  b.crd          = dim.crd() ;

  i_lo = b.crd * blockSize ;
  i_hi = min(i_lo + blockSize, extent) - 1 ;

  if(i_lo > i_hi)
    b.blockCount = 0 ;
  else
    blockParams(b, i_lo, i_hi) ;
}


// Iteration over all blocks of range.

void BlockRangeRep :: allBlocksEnum(Block& b) {
  int i_lo, i_hi ;

  b.crd = 0 ;

  i_lo = 0 ;
  i_hi = min(blockSize, extent) - 1 ;

  blockParams(b, i_lo, i_hi) ;
}

int BlockRangeRep :: allBlocksTest(Block& b) {return b.blockBase < extent ; }

void BlockRangeRep :: allBlocksNext(Block& b) {
  int i_lo, i_hi ;

  b.crd++ ;

  i_lo = b.blockBase + blockSize ;
  i_hi = min(i_lo + blockSize, extent) - 1 ;

  blockParams(b, i_lo, i_hi) ;
}


// Memory management

MEMORY_POOL_DEFN(BlockRangeRep)

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedBlockRangeRep': Block-distributed range yielding       //
//                              arrays with ``packed'' layout.         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

PackedBlockRangeRep :: PackedBlockRangeRep(int _extent, Dimension _dim,
                                           const int _blockSize) :
                     BlockRangeRep(_extent, _dim, _blockSize) {}

StrideRep* PackedBlockRangeRep :: subrngVector(const int dataLen,
                                               const int stride) {

  if(stride == 0) AdlibError("PackedBlockRangeRep :: subrngVector: "
			     "subrange stride is zero") ;

  if(stride > 0) {
    const int blockWidth = (blockSize + stride - 1) / stride ;

    return new PackedStrideRep(dataLen, stride, blockWidth) ;
  }
  else {
    const int blockWidth = (blockSize - stride - 1) / (-stride) ;

    return new PackedStrideRep(-dataLen, stride, blockWidth) ;
  }
}

MEMORY_POOL_DEFN(PackedBlockRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ExtendedBlockRangeRep': Block-distributed range with         //
//                                yielding arrays with ghost edges.    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

ExtendedBlockRangeRep :: ExtendedBlockRangeRep(int _extent, Dimension _dim,
                             const int _blockSize,
                             const int _wlo, const int _whi) :
                     BlockRangeRep(_extent, _dim, _blockSize),
                     wlo(_wlo), whi(_whi) {}

StrideRep* ExtendedBlockRangeRep :: vector(const int dataLen) {
  return new ExtendedStrideRep(dataLen, blockSize, wlo, whi) ;
}

MEMORY_POOL_DEFN(ExtendedBlockRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockSubRangeRep': range aligned to a                        //
//                           block-distributed range.                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

BlockSubRangeRep :: BlockSubRangeRep(const int _extent,
                        const int _base, const int _stride,
                        BlockRangeRep* _parent,
                        Dimension _dim, const int _blockSize) :
                            BlockRangeRep(_extent, _dim, _blockSize),
                            base(_base), stride(_stride),
                            parent(_parent) {
  parent->incr() ;
}

BlockSubRangeRep :: ~BlockSubRangeRep() {
  if(parent->decr() == 0) delete parent ;
}

RangeRep* BlockSubRangeRep :: prm() {return parent ; }
int BlockSubRangeRep :: bas() {return base ; }
int BlockSubRangeRep :: str() {return stride ; }

// Subscripting of Range.

Subscript* BlockSubRangeRep :: subscr(const int _i) {
  return BlockRangeRep :: subscr(base + stride * _i) ;
}

int BlockSubRangeRep :: subscr(const Subscript* s) {
  return (BlockRangeRep :: subscr(s) - base) / stride ;
}


// Creation of child ranges and strides...

RangeRep* BlockSubRangeRep :: subrng(const int _extent,
                                     const int _base, const int _stride) {
  return new BlockSubRangeRep(_extent,
                              base + stride * _base, stride * _stride,
                              parent, dim, blockSize) ;
}

StrideRep* BlockSubRangeRep :: vector(const int dataLen) {
  return parent->subrngVector(dataLen, stride) ;
}


// Auxilliary functions for use during block iteration

void BlockSubRangeRep :: blockInit(Block& b) {
  b.dim          = dim ;

  b.blockStep    = 1 ;

  b.blockStepSub = stride ;
}

void BlockSubRangeRep :: blockParams(Block& b, int i_lo, int i_hi) {
  b.blockCount   = i_hi - i_lo + 1 ;

  b.blockBase    = i_lo ;

  b.blockBaseSub = base + stride * i_lo - blockSize * b.crd ;
                       // (base + stride * i_lo) % blockSize
}


void BlockSubRangeRep :: blockRange(int& i_lo, int& i_hi, int _crd) {

  // Define `i_lo' and `i_hi' for current value of `_crd'

  // If `_crd' does not define a position within the bounds of
  // the range, the results are undefined but it is guaranteed that
  // i_hi < i_lo

  const int blockStart = _crd * blockSize ;

  const int blockStartOffset  = blockStart - base ;
  const int blockFinishOffset = blockStartOffset + blockSize - 1 ;

  if(stride > 0) {
    i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
    if(blockFinishOffset >= 0)
      i_hi = min(blockFinishOffset / stride, extent - 1) ;
    else
      i_hi = -1 ;
  }
  else {
    i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
    if(blockStartOffset <= 0)
      i_hi = min(blockStartOffset / stride, extent - 1) ;
    else
      i_hi = -1 ;
  }
}


// Iteration over locally held blocks.

void BlockSubRangeRep :: locBlockInit(Block& b) {
  int i_lo, i_hi ;

  b.crd          = dim.crd() ;

  blockRange(i_lo, i_hi, b.crd) ;

  if(i_lo > i_hi)
    b.blockCount = 0 ;
  else
    blockParams(b, i_lo, i_hi) ;
}


// Iteration over all blocks of range.

void BlockSubRangeRep :: allBlocksEnum(Block& b) {
  int i_lo, i_hi ;

  b.crd = base / blockSize ;

  blockRange(i_lo, i_hi, b.crd) ;

  if(i_lo > i_hi && i_lo < extent)
    allBlocksNext(b) ;
  else
    blockParams(b, i_lo, i_hi) ;
}

void BlockSubRangeRep :: allBlocksNext(Block& b) {
  int i_lo, i_hi ;

  do {
    if(stride > 0)
      b.crd++ ;
    else
      b.crd-- ;

    blockRange(i_lo, i_hi, b.crd) ;
  } while (i_lo > i_hi && i_lo < extent) ;

  blockParams(b, i_lo, i_hi) ;
}


// Memory management

MEMORY_POOL_DEFN(BlockSubRangeRep)

// -------------------------------------------------------------
// Things to do
//

