
// Operations on simple cyclic distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"


inline int max(int x, int y) {return x > y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicRangeRep': Simple-cyclic-distributed range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

CyclicRangeRep :: CyclicRangeRep(int _extent, Dimension _dim,
                                 const Layout _layout) :
                      RangeRep(_extent, _dim), layout(_layout) {
  numProcs = dim.size() ;
}

RangeRep* CyclicRangeRep :: prm() {return this ; }
int CyclicRangeRep :: bas() {return 0 ; }
int CyclicRangeRep :: str() {return 1 ; }

int CyclicRangeRep :: isScalar() {return 0 ; }


// Subscripting of Range.

Subscript* CyclicRangeRep :: subscr(const int _i) {
  crd = _i % numProcs ;
  sub = _i / numProcs ;

  return this ;
}

int CyclicRangeRep :: subscr(const Subscript* s) {
  return numProcs * s->sub + s->crd ;
}


// Creation of child ranges and strides...

RangeRep* CyclicRangeRep :: subrng(const int _extent,
                                   const int _base, const int _stride) {
  return new CyclicSubRangeRep(_extent, _base, _stride, this,
                               dim, layout) ;
}

StrideRep* CyclicRangeRep :: vector(const int dataLen) {
  const int volume = (extent + numProcs - 1) / numProcs ;

  return new SimpleStrideRep(dataLen, volume) ;
}


// Auxilliary functions for use during block iteration

void CyclicRangeRep :: blockInit(Block& b) {
  b.dim          = dim ;

  b.blockStep    = numProcs ;

  b.blockStepSub = 1 ;
}

void CyclicRangeRep :: blockParams(Block& b, int i_lo, int i_hi) {
  b.blockCount   = (i_hi - i_lo) / numProcs + 1 ;

  b.blockBase    = i_lo ;

  b.blockBaseSub = 0 ;
}


// Iteration over locally held blocks.

void CyclicRangeRep :: locBlockInit(Block& b) {
  int i_hi, i_lo ;

  b.crd          = dim.crd() ;

  i_lo = b.crd ;
  i_hi = b.crd + numProcs * ((extent - 1 - b.crd) / numProcs) ;

  if(extent <= b.crd)
    b.blockCount = 0 ;
  else
    blockParams(b, i_lo, i_hi) ;
}


// Iteration over all blocks of range.

void CyclicRangeRep :: allBlocksEnum(Block& b) {
  int i_hi, i_lo ;

  b.blk = 0 ;
  b.crd = 0 ;

  i_lo = 0 ;
  i_hi = numProcs * (extent - 1) / numProcs ;

  if(extent == 0)
    b.blk = 1 ;
  else
    blockParams(b, i_lo, i_hi) ;
}

int CyclicRangeRep :: allBlocksTest(Block& b) {return b.blk == 0 ; }

void CyclicRangeRep :: allBlocksNext(Block& b) {
  int i_hi, i_lo ;

  b.crd++ ;
  if(b.crd < numProcs) {
    i_lo = b.crd ;
    i_hi = b.crd + numProcs * ((extent - 1 - b.crd) / numProcs) ;

    if(extent <= b.crd)
      b.blk = 1 ;
    else
      blockParams(b, i_lo, i_hi) ;
  }
  else
    b.blk = 1 ;
}


// Memory management

MEMORY_POOL_DEFN(CyclicRangeRep)

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicSubRangeRep': subrange of a                            //
//                            simple-cyclic distributed range.         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


CyclicSubRangeRep :: CyclicSubRangeRep(const int _extent,
                         const int _base, const int _stride,
                         CyclicRangeRep* _parent,
                         Dimension _dim, const Layout _layout) :
			     CyclicRangeRep(_extent, _dim, _layout),
			     base(_base), stride(_stride), parent(_parent) {
  parent->incr() ;

  // Compute `f' and `f_s'.

  int g, g_s ;

  f   = numProcs ;
  f_s = 0 ;
  if(stride > 0) {
    g   = stride ;
    g_s = 1 ;
  }
  else {
    g   = -stride ;
    g_s = -1 ;
  }
  while(g) {
    int temp ;
    int r = f / g ;

    temp = g ;
    g    = f - r * g ;
    f    = temp ;
  
    temp = g_s ;
    g_s  = f_s - r * g_s ;
    f_s  = temp ;
  }

  i_step = numProcs / f ;
  b_step = stride / f ;
}

CyclicSubRangeRep :: ~CyclicSubRangeRep() {
  if(parent->decr() == 0) delete parent ;
}

RangeRep* CyclicSubRangeRep :: prm() {return parent ; }
int CyclicSubRangeRep :: bas() {return base ; }
int CyclicSubRangeRep :: str() {return stride ; }


// Subscripting of Range.

Subscript* CyclicSubRangeRep :: subscr(const int _i) {
  const int primIndex = base + stride * _i ;

  crd = primIndex % numProcs ;
  sub = primIndex / numProcs ;

  return this ;
}

int CyclicSubRangeRep :: subscr(const Subscript* s) {
  const int primIndex = numProcs * s->sub + s->crd ;

  return (primIndex - base) / stride ;
}


// Creation of child ranges and strides...

RangeRep* CyclicSubRangeRep :: subrng(const int _extent,
                                      const int _base, const int _stride) {
  return new CyclicSubRangeRep(_extent,
                               base + stride * _base, stride * _stride, parent,
                               dim, layout) ;
}

StrideRep* CyclicSubRangeRep :: vector(const int dataLen) {
  if(layout == STANDARD) {
    const int volume = (parent->size() + numProcs - 1) / numProcs ;

    return new SimpleStrideRep(dataLen, volume) ;
  }
  else {

    // layout == ADAPT

    const int truncPrimExtent =
        stride > 0 ? base + stride * (extent - 1) + 1 : base + 1 ;

    const int volume = (truncPrimExtent + numProcs - 1) / numProcs ;

    return new SimpleStrideRep(dataLen, volume) ;
  }
}


// Auxilliary functions for use during block iteration

void CyclicSubRangeRep :: blockInit(Block& b) {
  b.dim          = dim ;

  b.blockStep    = i_step ;

  b.blockStepSub = b_step ;
}

void CyclicSubRangeRep :: blockParams(Block& b, int i_lo, int i_hi) {
  b.blockCount   = (i_hi - i_lo) / i_step + 1 ;

  b.blockBase    = i_lo ;

  b.blockBaseSub = (base + stride * i_lo) / numProcs ;
}

void CyclicSubRangeRep :: blockRange(int& i_lo, int& i_hi, int _crd) {

  // Define `i_lo' and `i_hi' for current value of `_crd'

  // Compute the smallest and largest solutions of
  //
  //     base + i * s = crd    mod  p
  //
  // in the range 0, ..., extent - 1, viz
  //
  //     i_lo = a + b * ceiling(- a / b)
  //     i_hi = a + b * floor((extent - 1 - a) / b)
  //
  // where a = c / f * f_s, b = p / f, c = crd - base.
  //
  // Solutions exist iff `f' divides `c'.

  int c = _crd - base ;
  if(c % f == 0) {
    int a = (c / f) * f_s, b = numProcs / f ;

    if(a >= 0)
      i_lo = a - b * (a / b) ;
    else
      i_lo = a + b * ((-a + b - 1) / b) ;

    if(extent > a)
      i_hi = a + b * ((extent - 1 - a) / b) ;
    else
      i_hi = a - b * ((-extent + a + b) / b) ;
  }
  else {
    i_lo =  0 ;
    i_hi = -1 ;
  }
}


// Iteration over locally held blocks.

void CyclicSubRangeRep :: locBlockInit(Block& b) {
  int i_lo, i_hi ;

  b.crd          = dim.crd() ;

  blockRange(i_lo, i_hi, b.crd) ;

  if(i_lo > i_hi)
    b.blockCount = 0 ;
  else
    blockParams(b, i_lo, i_hi) ;
}


// Iteration over all blocks of range.

void CyclicSubRangeRep :: allBlocksEnum(Block& b) {
  int i_lo, i_hi ;

  b.crd = base % numProcs ;
  blockRange(i_lo, i_hi, b.crd) ;

  if(i_lo > i_hi)   // Should only happen if extent = 0
    b.blk = 1 ;
  else {
    b.blk = 0 ;
    blockParams(b, i_lo, i_hi) ;
  }
}

int CyclicSubRangeRep :: allBlocksTest(Block& b) {return b.blk == 0 ; }

void CyclicSubRangeRep :: allBlocksNext(Block& b) {
  int i_lo, i_hi ;

  b.crd = (b.crd + stride) % numProcs ;
  if(b.crd < 0) b.crd += numProcs ;  // because `stride' may be -ve.

  if(b.crd != base % numProcs) {
    blockRange(i_lo, i_hi, b.crd) ;

    if(i_lo > i_hi)  // Should only happen if i > extent
      b.blk = 1 ;
    else
      blockParams(b, i_lo, i_hi) ;
  }
  else  // Have returned to starting position.
    b.blk = 1 ;
}


// Memory management

MEMORY_POOL_DEFN(CyclicSubRangeRep)

// ------------------------------------------------------------------
// Things to do
//
//   Shift computation of `f' and `f_s' into `blockInit'?
//   Not entirely natural, but the current direction is towards the
//   `Range' construction being a light-weight operation.
//
//   `blk'.
