
// Schedules for irregular subscripting of distributed arrays.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"
#include "threads.h"
#include "ProcTree.h"
#include "Procs.h"
#include "Range.h"
#include "SFrame.h"
#include "Frame.h"
#include "Seq.h"
//#include "SArray.h"
//#include "Array.h"
#include "schedule.h"
#include "MessSchedule.h"
#include "TreeSchedule.h"
#include "DataSchedule.h"
#include "Reduce.h"
#include "Gather.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `GatherSkeleton'                                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

GatherSkeleton :: GatherSkeleton(const Frame* _loc, const Frame* _rem,
                                 const int len,
                                 const Frame** _subs, int** subsData,
                                 const Frame* _msk, int* mskData) :
      DataSchedule(len),
      loc(_loc), rem(_rem), subs(_subs), msk(_msk), rank(_loc->rnk()) {

  Group p = loc->grp() ;
  if(p.on()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.
    rep.where() ;
      if(apg.lead() == p.lead())
	loop(0, subsData, mskData, 0) ;
    rep.erewh() ;
  } p.no() ;

  TreeSchedule spanTree ;
  apg.spanTree(&spanTree) ;
  DataSchedule :: build(&spanTree) ;
}

void GatherSkeleton :: loop(int locOffset, int** subsElem, int* mskElem,
                            const int r) {

  if(r == rank) {
    if(!msk || *mskElem) {
      Group  remGrp = rem->grp() ;
      int remOffset = 0 ;

      for(int s = 0 ; s < rem->rnk() ; s++) {
	Range y(rem->rng(s)) ;
	Subscript& i = y(*(subsElem [s])) ;

        remOffset += rem->str(s).offset(i) ;
        remGrp    /= i ;
      }

      transReq(locOffset, remGrp.lead(), remOffset) ;
    }
  }
  else {
    LocBlocksIndex i(loc->rng(r)) ;

    Stride locStr = loc->str(r) ;
    Stride mskStr ;
    if(msk) mskStr = msk->str(r) ;

    int** subsDataElem = new int* [rem->rnk()] ;
    int* subsStep = new int [rem->rnk()] ;

    for(i.begin() ; i.test() ; i.next()) {
      i.sub = i.blockBaseSub ;

      int s ;

      int locOffsetElem = locOffset + locStr.offset(i) ;
      const int locStep = locStr.step(i) ;

      for(s = 0 ; s < rem->rnk() ; s++) {
        Stride subStr = (subs [s])->str(r) ;

        subsDataElem [s] = subsElem [s] + subStr.offset(i) ;
        subsStep [s]     = subStr.step(i) ;
      }

      int* mskDataElem  = msk ? mskElem + mskStr.offset(i) : (int*) NULL ;
      const int mskStep = msk ? mskStr.step(i) : 0 ;

      for(int i_block = 0 ; i_block < i.blockCount ; i_block++) {
        loop(locOffsetElem, subsDataElem, mskDataElem, r + 1) ;

	locOffsetElem += locStep ;
        for(s = 0 ; s < rem->rnk() ; s++)
          subsDataElem [s] += subsStep [s] ;
        if(msk)
          mskDataElem += mskStep ;
      }
    }

    delete [] subsStep ;
    delete [] subsDataElem ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ReshapeSkeleton'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

ReshapeSkeleton :: ReshapeSkeleton(const Frame* _loc, const Frame* _rem,
                                   const int len) :
		       loc(_loc), rem(_rem), rank(_loc->rnk()),
		       DataSchedule(len) {

  Group p = loc->grp() ;
  if(p.on()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.
    rep.where() ;
      if(apg.lead() == p.lead())
        loop(0, 0, 1, 0) ;
    rep.erewh() ;
  } p.no() ;

  TreeSchedule spanTree ;
  apg.spanTree(&spanTree) ;
  DataSchedule :: build(&spanTree) ;
}


void ReshapeSkeleton :: loop(int locOffset, int seqIndex, int seqStride,
                             int r) {

  if(r == rank) {
    Group remGrp = rem->grp() ;
    int remOffset = 0 ;

    for(int s = 0 ; s < rem->rnk() ; s++) {
      Range y(rem->rng(s)) ;
      Subscript& i = y(seqIndex % y.size()) ;

      remOffset += rem->str(s).offset(i) ;
      remGrp    /= i ;
    }

    transReq(locOffset, remGrp.lead(), remOffset) ;
  }
  else {
    Range x = loc->rng(r) ;
    Index i(x) ;

    for(i.where() ; i.test() ; i.next()) {
      int locOffsetElem = locOffset + loc->str(r).offset(i) ;

      loop(locOffsetElem,
           seqIndex + seqStride * i, x.size() * seqStride,
	   r + 1) ;
    } i.erewh() ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ReshapePadSkeleton'                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

ReshapePadSkeleton :: ReshapePadSkeleton(const Frame* _loc, int _remSize,
                                         const Frame* _pad,
                                         const int len) :
			  loc(_loc), remSize(_remSize), pad(_pad),
                          rank(_loc->rnk()),
			  DataSchedule(len) {

  // Silly `reshape' padding to support Fortran.

  Group p = loc->grp() ;
  if(p.on()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.
    rep.where() ;
      if(apg.lead() == p.lead())
	loop(0, 0, 1, 0) ;
    rep.erewh() ;
  } p.no() ;

  TreeSchedule spanTree ;
  apg.spanTree(&spanTree) ;
  DataSchedule :: build(&spanTree) ;
}

void ReshapePadSkeleton :: loop(int locOffset, int seqIndex, int seqStride,
                                int r) {

  if(r == rank) {
    if(seqIndex >= remSize) {
      Range y(pad->rng(0)) ;
      const int padIndex = (seqIndex - remSize) % y.size() ;
 
      Subscript& i = y(padIndex) ;

      Group remGrp = pad->grp() / i ;
      transReq(locOffset, remGrp.lead(), pad->str(0).offset(i)) ;
    }
  }
  else {
    Range x = loc->rng(r) ;
    Index i(x) ;

    for(i.where() ; i.test() ; i.next()) {
      loop(locOffset + loc->str(r).offset(i),
	   seqIndex + seqStride * i, x.size() * seqStride,
	   r + 1) ;
    } i.erewh() ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGatherScatterSkeleton'                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


int VecGatherSkeleton :: blockRank(const int rank, const Frame** subs) {
  int result = 0 ;
  for (int s = 0 ; s < rank ; s++)
    if(!(subs [s])) result++ ;

  return result ;
}

VecGatherSkeleton :: VecGatherSkeleton(const Frame* _loc, const Frame* _rem,
                                       const int len,
                                       const Frame** _subs, int** _subsData) :
                         BlockDataSchedule(len,
                                           blockRank(_rem->rnk(), _subs)),
                         loc(_loc), rem(_rem),
                         subs(_subs), subsData(_subsData),
                         rank(_loc->rnk()) {
  Group p = loc->grp() ;
  if(p.on()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.
    rep.where() ;
      if(apg.lead() == p.lead()) {
	exts     = new int [BlockDataSchedule :: rank] ;
	locSteps = new int [BlockDataSchedule :: rank] ;
	remSteps = new int [BlockDataSchedule :: rank] ;

	loop(0, 1, 0, 1, 1, rem->grp(), 0, 0) ;

	delete [] remSteps ;
	delete [] locSteps ;
	delete [] exts ;
      }
    rep.erewh() ;
  } p.no() ;

  TreeSchedule spanTree ;
  apg.spanTree(&spanTree) ;
  BlockDataSchedule :: build(&spanTree) ;
}

void VecGatherSkeleton :: loop(const int locOffset, const int locCtg,
                               const int remOffset, const int remCtg,
                               const int vol,
                               Group remGrp,
                               const int r, const int s) {

  if(r == rank)
    transReq(locOffset, locSteps, locCtg, remOffset, remSteps, remCtg,
             exts, vol, remGrp.lead()) ;
  else {

    Range  x      = loc->rng(r) ;
    Stride locStr = loc->str(r) ;

    Range  y      = rem->rng(r) ;
    Stride remStr = rem->str(r) ;

    const Frame* sub = subs [r] ;

    if(sub) {

      // Dimension has a vector subscript.

      Index i(x) ;

      for(i.where() ; i.test() ; i.next()) {
        int* subData = subsData [r] ;
        Subscript& j = y(subData [sub->str(0).offset(i)]) ;

	loop(locOffset + locStr.offset(i), locCtg,
             remOffset + remStr.offset(j), remCtg,
             vol,
             remGrp / j, r + 1, s) ;
      } i.erewh() ;
    }
    else {

      // Dimension is unsubscripted.  (cf `remap').

      LocBlocksIndex i(x) ;

      for(i.begin() ; i.test() ; i.next()) {
	i.sub = i.blockBaseSub ;

	int locOffsetElem = locOffset + locStr.offset(i) ;
	const int locStep = locStr.step(i) ;

        AllBlocksIndex j(y.subrng(i.blockCount, i.blockBase, i.blockStep)) ;

	for(j.begin() ; j.test() ; j.next()) {
   	  j.sub = j.blockBaseSub ;

	  const int remOffsetElem = remOffset + remStr.offset(j) ;
	  const int remStep       = remStr.step(j) ;

          const int jlocStep = locStep * j.blockStep ;

	  exts     [s] = j.blockCount ;
          locSteps [s] = jlocStep ;
	  remSteps [s] = remStep ;

	  loop(locOffsetElem + j.blockBase * locStep, locCtg && jlocStep == vol,
               remOffsetElem, remCtg && remStep == vol,
               vol * j.blockCount,
               remGrp / j,
               r + 1, s + 1) ;
	}
      }
    }
  }
}


// -------------------------------------------------------------
// Things to do
//
//   Unify `Reshape' and `ReshapePad' by introducing a start offset in former?
//
//   Could `reshape' be generalized to support arbitrary F77 ``sequence
//       association'', somehow?
//
//   In any case, `reshape' should be based on a `MessSchedule'...
//       it is ``deterministic''.
//
//   Add a version of `VecGatherSkeleton' that works on a `DataSchedule',
//       requiring all dimensions to be vector subscripted.
//
//   Eliminate use of `Index' iterators in in `VecGatherSkeleton' and
//       `Reshape' (in favour of lower-level `LocBlocksIndex').
//
//   Where there are several short `int' vectors in a schedule (`exts',
//     `steps', ...) allocate and deallocate them as a single vector.

