
enum Mode {CYCL, EDGE} ;

class RemapSkeleton : public BlockMessSchedule {
public :
  RemapSkeleton(const Frame* dst, const Frame* src, const int len) ;
private :
  const Frame* rem ;
  const Frame* loc ;

  int rank ;

  int* exts ;
  int* steps ;

  void (RemapSkeleton :: *transfer)(int, int, int, int) ;

  void loop(const int offset, const int ctg, const int vol,
            Group remGrp, const int r) ;

  void send(int offset, int ctg, int vol, int id) ;
  void recv(int offset, int ctg, int vol, int id) ;
} ;

class ShiftSkeleton : public BlockMessSchedule {
public :
  ShiftSkeleton(const Frame* dst, const Frame* src, const int len,
                const int shift, const int dim, const Mode mode) ;
private :
  int rank, dim ;

  int nShfts ;
  int shftExts [2] ;
  int shftSrcDisps [2] ;
  int shftDstDisps [2] ;

  int* exts ;
  int* steps ;

  const Frame* loc ;

  int* shftRemDisps ;
  int* shftLocDisps ;

  void (ShiftSkeleton :: *transfer)(int, int, int, int) ;

  void loop(const int offset, const int ctg, const int vol,
            Group remGrp, const int r) ;

  void send(int offset, int ctg, int vol, int id) ;
  void recv(int offset, int ctg, int vol, int id) ;
} ;


class SkewSkeleton : public BlockMessSchedule {
public :
  SkewSkeleton(const Frame* _dst, const Frame* _src, const int len,
               const Frame* _shf, int* shfDat,
               const int _dim, const Mode _mode) ;
private :
  const Frame* dst ;
  const Frame* src ;
  const Frame* shf ;

  int dim ;
  Mode mode ;

  int rank ;

  void loop(const int dstOffset, const int srcOffset, int* shfData,
            const int r, const int s) ;

  void shift(int offset, Stride str,
             const int ext, const int locDisp, const int remDisp,
             void (SkewSkeleton :: *transfer)(int, int, int, int)) ;

  void send(int offset, int step, int count, int id) ;
  void recv(int offset, int step, int count, int id) ;
} ;

class MultiShiftSkeleton : public BlockMessSchedule {
public :
  MultiShiftSkeleton(const Frame* dst, const Frame* src, const int len,
                     int* shift, Mode* mode) ;
private :
  int rank ;

  int* nShfts ;
  int (*shftExts) [2] ;
  int (*shftSrcDisps) [2] ;
  int (*shftDstDisps) [2] ;

  int* exts ;
  int* steps ;

  const Frame* loc ;

  int (*shftRemDisps) [2] ;
  int (*shftLocDisps) [2] ;

  void (MultiShiftSkeleton :: *transfer)(int, int, int, int) ;

  void loop(const int offset, const int ctg, const int vol,
            Group remGrp, const int r) ;

  void send(int offset, int ctg, int vol, int id) ;
  void recv(int offset, int ctg, int vol, int id) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `WriteFaces'                                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class WriteFaces : public BlockMessSchedule {
public :
  WriteFaces(const Frame* _src, const int len,
             const int _dim, const int* _wlo, const int* _whi) ;
private :
  const Frame* src ;

  int rank, dim ;
  const int* wlo ;
  const int* whi ;

  int* exts ;
  int* steps ;

  void loop(const int offset, const int ctg, const int vol,
            const Group& remGrp, const int r) ;

  void recvLoop(const int offset, const int ctg, const int vol,
                const Group& remGrp, const int r) ;

  void sendLoop(const int offset, const int ctg, const int vol,
                const Group& remGrp, const int r) ;

} ;



class Remap {
public :
  Remap(const Frame* dst, const Frame* src, const int len) :
          skeleton(dst, src, len), bcst(dst, len) {}

  void execute(void* dstDat, void* srcDat) {
    skeleton.gather(dstDat, srcDat) ;
    bcst.execute(dstDat) ;
  }
private :
  RemapSkeleton skeleton ;
  Replicate bcst ;
} ;


class Shift {
public :
  Shift(const Frame* dst, const Frame* src, const int len,
        const int shift, const int dim, const Mode mode) :
          skeleton(dst, src, len, shift, dim, mode) {}

  void execute(void* dstDat, void* srcDat) {
    skeleton.gather(dstDat, srcDat) ;
  }
private :
  ShiftSkeleton skeleton ;
} ;

class Skew {
public :
  Skew(const Frame* dst, const Frame* src, const int len,
       const Frame* shf, int* shfDat, const int dim, const Mode mode) :
          skeleton(dst, src, len, shf, shfDat, dim, mode) {}

  void execute(void* dstDat, void* srcDat) {
    skeleton.gather(dstDat, srcDat) ;
  }
private :
  SkewSkeleton skeleton ;
} ;

class MultiShift {
public :
  MultiShift(const Frame* dst, const Frame* src, const int len,
             int* shift, Mode* mode) :
          skeleton(dst, src, len, shift, mode) {}

  void execute(void* dstDat, void* srcDat) {
    skeleton.gather(dstDat, srcDat) ;
  }
private :
  MultiShiftSkeleton skeleton ;
} ;


class WriteHalo {
public :
  WriteHalo(const Frame* src, const int len, const int* wlo, const int* whi) :
          rank(src->rnk()) {
    faceSchedules = new WriteFaces* [rank] ;
    for(int r = 0 ; r < rank ; r++)
      if(wlo [r] > 0 || whi [r] > 0)
	faceSchedules [r] = new WriteFaces(src, len, r, wlo, whi) ;
      else
	faceSchedules [r] = 0 ;
  }

  ~WriteHalo() {
    for(int r = 0 ; r < rank ; r++)
      if(faceSchedules [r]) delete faceSchedules [r] ;
    delete [] faceSchedules ;
  }

  void execute(void* base) {
    for(int r = 0 ; r < rank ; r++)
      if(faceSchedules [r]) faceSchedules [r]->gather(base, base) ;
  }
private :
  const int rank ;
  WriteFaces** faceSchedules ;
} ;


// -------------------------------------------------------------
// Things to do
//
//   Make `Group' arguments to `loop' into `const Group &'.
//
//   Do we really need separate `Shift'/`ShiftSkeleton', etc classes?

