
// Operations on collapsed ranges.

#include <stdlib.h>
#include <stdio.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CollapsedRangeRep': Collapsed range of distributed arrray.   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

CollapsedRangeRep :: CollapsedRangeRep(const int _extent,
                                       const Layout _layout) :
                         RangeRep(_extent, Dimension()), layout(_layout) {}

RangeRep* CollapsedRangeRep :: prm() {return this ; }
int CollapsedRangeRep :: bas() {return 0 ; }
int CollapsedRangeRep :: str() {return 1 ; }

//int CollapsedRangeRep :: isScalar() {return 0 ; }


// Subscripting of Range.

Subscript* CollapsedRangeRep :: subscr(const int _i) {
  blk = 0  ;
  sub = _i ;

  return this ;
}

int CollapsedRangeRep :: subscr(const Subscript* s) {
  return s->sub ;
}


// Creation of child ranges and strides...

RangeRep* CollapsedRangeRep :: subrng(const int _extent,
                                      const int _base, const int _stride) {
  return new CollapsedSubRangeRep(_extent, _base, _stride, this, layout) ;
}

StrideRep* CollapsedRangeRep :: vector(const int dataLen) {
  return new SimpleStrideRep(dataLen, extent) ;
}

// Auxilliary functions for use during block iteration

void CollapsedRangeRep :: blockInit(Block& b) {
  b.dim          = dim ;

  b.blockStep    = 1 ;

  b.blockStepSub = 1 ;
}

// Iteration over locally held blocks.

void CollapsedRangeRep :: locBlockInit(Block& b) {
  if(extent == 0)
    b.blockCount = 0 ;
  else {
    b.blockCount   = extent ;

    b.blockBase    = 0 ;

    b.blockBaseSub = 0 ;
  }
}


// Iteration over all blocks of range.

void CollapsedRangeRep :: allBlocksEnum(Block& b) {
  locBlockInit(b) ;
  locBlocksEnum(b) ;
}

int CollapsedRangeRep :: allBlocksTest(Block& b) {return locBlocksTest(b) ; }

void CollapsedRangeRep :: allBlocksNext(Block& b) {locBlocksNext(b) ; }


// Memory management

MEMORY_POOL_DEFN(CollapsedRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CollapsedSubRangeRep': subrange of a collapsed range.        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


CollapsedSubRangeRep :: CollapsedSubRangeRep(const int _extent,
                            const int _base, const int _stride,
                            CollapsedRangeRep* _parent,
                            const Layout _layout) :
				CollapsedRangeRep(_extent, _layout),
				base(_base), stride(_stride),
				parent(_parent) {
  parent->incr() ;
}

CollapsedSubRangeRep :: ~CollapsedSubRangeRep() {
  if(parent->decr() == 0) delete parent ;
}


RangeRep* CollapsedSubRangeRep :: prm() {return parent ; }
int CollapsedSubRangeRep :: bas() {return base ; }
int CollapsedSubRangeRep :: str() {return stride ; }


// Subscripting of Range.

Subscript* CollapsedSubRangeRep :: subscr(const int _i) {
  blk = 0 ;
  sub = base + stride * _i ;

  return this ;
}

int CollapsedSubRangeRep :: subscr(const Subscript* s) {
  return (s->sub - base) / stride ;
}


// Creation of child ranges and strides...

RangeRep* CollapsedSubRangeRep :: subrng(const int _extent,
                                         const int _base, const int _stride) {
  return new CollapsedSubRangeRep(_extent,
                                  base + stride * _base, stride * _stride,
                                  parent,
                                  layout) ;
}

StrideRep* CollapsedSubRangeRep :: vector(const int dataLen) {

  if(layout == STANDARD)
    return new SimpleStrideRep(dataLen, parent->size()) ;
  else {
    // layout == PACKED

    if(stride == 0) AdlibError("CollapsedSubRangeRep :: vector: "
			       "range stride is zero") ;

    if(stride > 0) {
      const int blockWidth = (parent->size() + stride - 1) / stride ;

      return new PackedStrideRep(dataLen, stride, blockWidth) ;
    }
    else {
      const int blockWidth = (parent->size() - stride - 1) / (-stride) ;

      return new PackedStrideRep(-dataLen, stride, blockWidth) ;
    }
  }
}


// Auxilliary functions for use during block iteration

void CollapsedSubRangeRep :: blockInit(Block& b) {
  b.dim          = dim ;

  b.blockStep    = 1 ;

  b.blockStepSub = stride ;
}

// Iteration over locally held blocks.

void CollapsedSubRangeRep :: locBlockInit(Block& b) {
  if(extent == 0)
    b.blockCount = 0 ;
  else {
    b.blockCount   = extent ;

    b.blockBase    = 0 ;

    b.blockBaseSub = base ;
  }
}


// Memory management

MEMORY_POOL_DEFN(CollapsedSubRangeRep)

