
// Schedules for deterministic communication

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"  // debug
#include "Procs.h"
#include "threads.h"
#include "adtags.h"
#include "schedule.h"
#include "MessSchedule.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `MessSchedule' class.                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

MessSchedule :: MessSchedule(const int _elementLen) : elementLen(_elementLen) {
  MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;
  MPI_Comm_size(MPI_COMM_WORLD, &nProcs) ;

  locIdxs = new IndexList [nProcs] ;
  srvIdxs = new IndexList [nProcs] ;
}

void MessSchedule :: build() {

  // Set `locList' and `srvList' to hop-ordered lists of non-empty messages,
  // and allocate contiguous buffers big enough to hold largest messages.

  int locBufSize = 0, srvBufSize = 0 ;
  for(int hop = 0 ; hop < nProcs ; hop++) {
    {
      const int remId = (locId + hop) % nProcs ;
      int size = locIdxs [remId].size() ;
      if(size) {
	locList.add(remId) ;
        if(size > locBufSize) locBufSize = size ;
      }
    }
    {
      const int remId = (locId - hop + nProcs) % nProcs ;
      int size = srvIdxs [remId].size() ;
      if(size) {
	srvList.add(remId) ;
        if(size > srvBufSize) srvBufSize = size ;
      }
    }
  }
  locBuf = new char [locBufSize * elementLen] ;
  srvBuf = new char [srvBufSize * elementLen] ;
}

MessSchedule :: ~MessSchedule() {
  delete [] srvBuf ;
  delete [] locBuf ;

  delete [] srvIdxs ;
  delete [] locIdxs ;
}

void MessSchedule :: gather(void* locBase, void* srvBase) {

  // Spawn server to send messages in `srvList'...

  SendServer sndServer(srvList.size(), srvList.data(),
                       srvBuf, srvIdxs, srvBase, elementLen) ;

  // Receive messages in `locList'...

  const int numMes = locList.size() ;
  int* remIds      = locList.data() ;
  for(int mes = 0 ; mes < numMes ; mes++) {
    const int remId = remIds [mes] ;

    IndexList* indexVec = locIdxs + remId ;
    const int count = indexVec->size() ;
    int*    offsets = indexVec->data() ;

    MPI_Status status ;
    MPI_Irecv(locBuf, count * elementLen, MPI_CHAR,
              remId, DAT, MPI_COMM_WORLD, event) ;
    Server :: waitClientEvent(&status) ;

    scatterOffBuf((char*) locBase, offsets, count, locBuf, elementLen) ;
  }

  // Wait for `sndServer' to terminate.

  while(sndServer.active())
    Server :: waitServerTransition() ;
}

void MessSchedule :: scatter(void* locBase, void* srvBase) {

  // Spawn server to send messages in `locList'...

  SendServer sndServer(locList.size(), locList.data(),
                       locBuf, locIdxs, locBase, elementLen) ;

  // Receive messages in `srvList'...

  const int numMes = srvList.size() ;
  int* remIds      = srvList.data() ;
  for(int mes = 0 ; mes < numMes ; mes++) {
    const int remId = remIds [mes] ;

    IndexList* indexVec = srvIdxs + remId ;
    const int count  = indexVec->size() ;
    int* offsets     = indexVec->data() ;

    MPI_Status status ;
    MPI_Irecv(srvBuf, count * elementLen, MPI_CHAR,
              remId, DAT, MPI_COMM_WORLD, event) ;
    Server :: waitClientEvent(&status) ;

    scatterOffBuf((char*) srvBase, offsets, count, srvBuf, elementLen) ;
  }

  while(sndServer.active())
    Server :: waitServerTransition() ;
}

void MessSchedule :: combine(void* locBase, void* srvBase,
                             void (*op)(void*, void*)) {

  // Spawn server to send messages in `locList'...

  SendServer sndServer(locList.size(), locList.data(),
                       locBuf, locIdxs, locBase, elementLen) ;

  // Receive messages in `srvList'...

  const int numMes = srvList.size() ;
  int* remIds      = srvList.data() ;
  for(int mes = 0 ; mes < numMes ; mes++) {
    const int remId = remIds [mes] ;

    IndexList* indexVec = srvIdxs + remId ;
    const int count  = indexVec->size() ;
    int* offsets     = indexVec->data() ;

    MPI_Status status ;
    MPI_Irecv(srvBuf, count * elementLen, MPI_CHAR,
              remId, DAT, MPI_COMM_WORLD, event) ;
    Server :: waitClientEvent(&status) ;

    char* buf = srvBuf ;
    for(int i = 0 ; i < count ; i++) {
      (*op)((char*) srvBase + elementLen * offsets [i], buf) ;
      buf += elementLen ;
    }
  }

  while(sndServer.active())
    Server :: waitServerTransition() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `SendServer' class.                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

SendServer :: SendServer(const int _numMes, int* _remIds,
			 char* _sendBuf, IndexList* _sendIdxs,
			 void* _sendBase, int _elementLen) :
                     numMes(_numMes), remIds(_remIds),
                     sendBuf(_sendBuf), sendIdxs(_sendIdxs),
                     sendBase(_sendBase), elementLen(_elementLen) {
  mes = 0 ;
  if(mes < numMes) initSend(remIds [mes]) ;
}

SendServer :: ~SendServer() {}

void SendServer :: transition(MPI_Status* status) {
  mes++ ;
  if(mes < numMes) initSend(remIds [mes]) ;
}

int SendServer :: active() const {
  return mes < numMes ;
}

void SendServer :: initSend(const int remId) {
  IndexList* indexVec = sendIdxs + remId ;
  const int count = indexVec->size() ;
  int*    offsets = indexVec->data() ;

  gatherOffBuf((char*) sendBase, offsets, count, sendBuf, elementLen) ;

  MPI_Isend(sendBuf, count * elementLen, MPI_CHAR,
	    remId, DAT, MPI_COMM_WORLD, event) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockMessSchedule' class.                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void BlockMessSchedule :: run(List<MessQueue>* dstMessages,
                              List<MessQueue>* srcMessages,
                              MultiBlockList dstCopies,
                              MultiBlockList srcCopies,
                              void* dstBase, void* srcBase) {

  // Spawn server to send messages in `srcMessages'...

  BlockSendServer sndServer(srcMessages, srcBase, elementLen) ;

  // Do local copying...

  MultiBlock dstBlock = dstCopies.data() ;
  MultiBlock srcBlock = srcCopies.data() ;

  for(int i = 0 ; i < cpyNum ; i++) {
    copyMblkMblk(dstBase, dstBlock, srcBase, srcBlock, elementLen) ;
    nextBlock(dstBlock) ;
    nextBlock(srcBlock) ;
  }

  // Receive messages in `dstList'...

  List<MessQueue>* record = dstMessages ;

  while(record) {
    const int  remId = record->remId ;
    const int numMes = record->messages.size() ;

    Mess*    message = record->messages.data() ;

    MultiBlock block = record->blocks.data() ;

    MPI_Status status ;

    for(int mes = 0 ; mes < numMes ; mes++) {

      const int num  = message->num  ;
      const int size = message->size ;

      if(num == 1 && block [MBLK_CTG]) {
        char* blockBase  = (char*) dstBase + elementLen * block [MBLK_OFF] ;

	MPI_Irecv(blockBase, size, MPI_CHAR,
                  remId, DAT, MPI_COMM_WORLD, event) ;
	Server :: waitClientEvent(&status) ;

        nextBlock(block) ;
      }
      else {
	MPI_Irecv(statBuf, size, MPI_CHAR,
                  remId, DAT, MPI_COMM_WORLD, event) ;
	Server :: waitClientEvent(&status) ;

        char* ptr = statBuf ;

        for(int j = 0 ; j < num ; j++) {
          scatterMblkBuf(dstBase, block, ptr, elementLen) ;

          ptr += block [MBLK_SIZ] ;
          nextBlock(block) ;
        }
      }

      message++ ;
    }
    
    record = record->next ;
  }

  // Wait for `sndServer' to terminate.

  while(sndServer.active())
    Server :: waitServerTransition() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockSendServer' class.                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

BlockSendServer :: BlockSendServer(List<MessQueue>* sendList,
                                   void* _sendBase, int _elementLen) :
		         record(sendList),
			 sendBase(_sendBase),
                         elementLen(_elementLen) {
  if(record) {
    remId   = record->remId ;
    numMes  = record->messages.size() ;

    mes     = 0 ;
    message = record->messages.data() ;

    block   = record->blocks.data() ;

    initSend() ;
  }
}

void BlockSendServer :: transition(MPI_Status* status) {
  mes++ ;
  if(mes < numMes) {
    message++ ;

    initSend() ;
  }
  else {
    record = record->next ;

    if(record) {
      remId   = record->remId ;
      numMes  = record->messages.size() ;

      mes     = 0 ;
      message = record->messages.data() ;

      block   = record->blocks.data() ;

      initSend() ;
    }
  }
}

void* BlockSendServer :: active() const {
  return record ;
}

void BlockSendServer :: initSend() {
  const int num  = message->num  ;
  const int size = message->size ;

  if(num == 1 && block [MBLK_CTG]) {
    char* blockBase  = (char*) sendBase + elementLen * block [MBLK_OFF] ;
    nextBlock(block) ;

    MPI_Isend(blockBase, size, MPI_CHAR,
	      remId, DAT, MPI_COMM_WORLD, event) ;
  }
  else {
    char* ptr = sendBuf ;

    for(int j = 0 ; j < num ; j++) {
      gatherMblkBuf(sendBase, block, ptr, elementLen) ;

      ptr += block [MBLK_SIZ] ;
      nextBlock(block) ;
    }

    MPI_Isend(sendBuf, size, MPI_CHAR,
	      remId, DAT, MPI_COMM_WORLD, event) ;
  }
}


// ----------------------------------------------------------------------
// Things to do
//
//     Change `MessSchedule' to use statically declared buffer.
//     General reorganizations to look more like modern `BlockMessSchedule'.
//
//     Replace `scatterOffBuf' by a virtual function, and unify implementations
//     of `gather', `scatter', `combine'.
//
//     Could `run' be a global function (as it was in the bad old
//     days), or a member of some class other than `BlockDataSchedule'?

