
// Process arrays and groups

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"
#include "ProcTree.h"
#include "Procs.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `ActiveProcessGroup': the process group executing this     //
//                                control thread.                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void ActiveProcessGroup :: init(Procs* _procs) {
  procs   = _procs ;
  mask    = procs->mask ; ;
  leadId  = 0 ;
  onLevel = 1 ;
}


void ActiveProcessGroup :: deactivate(DimensionRep* dim) {

#ifndef CHECK_STRONG_ALIGNMENT
  dim = procs->dim(dim->axis).rep ;
#endif

  Group :: deactivate(dim, dim->coord) ;
  dim->inactive = onLevel ;
}

void ActiveProcessGroup :: reactivate(DimensionRep* dim) {

#ifndef CHECK_STRONG_ALIGNMENT
  dim = procs->dim(dim->axis).rep ;
#endif

  dim->inactive = 0 ;
  Group :: reactivate(dim, dim->coord) ;
}


int ActiveProcessGroup :: on(Procs* p) {
  if(procs != p)
    if(p->parent.procs == procs && p->parent.mask == mask) {
      p->parent.onLevel = onLevel ;  // Cache current `onLevel' in `p->parent'.

      procs   = p ;
      mask    = p->mask ;
      leadId  = 0 ;
      onLevel = 0 ;
    }
    else
      AdlibError("ActiveProcessGroup :: on: inaccessible process array") ;
  onLevel++;

  return p->loc() ;
}

void ActiveProcessGroup :: no(Procs* p) {
  onLevel-- ;
  if(onLevel == 0) {
    onLevel = procs->parent.onLevel ;
    mask    = procs->parent.mask    ;
    leadId  = procs->parent.leadId  ;
    procs   = procs->parent.procs   ;
  }
}


int ActiveProcessGroup :: on(const Group* p) {
  int onFlag = on(p->procs) ;

  for(int i = 0 ; i < procs->rnk() ; i++) {
    DimensionRep* d = procs->dim(i).rep ;
    if(!(p->mask & d->mask) && !d->inactive)
      deactivate(d) ;
  }

  return onFlag && (lead() == p->lead()) ;
}

void ActiveProcessGroup :: no(const Group* p) {
  for(int i = 0 ; i < procs->rnk() ; i++) {
    DimensionRep* d = procs->dim(i).rep ;
    if(d->inactive == onLevel)
      reactivate(d) ;
  }

  no(p->procs) ;
}

void ActiveProcessGroup :: where(DimensionRep* d) {
  onLevel++ ;
  deactivate(d) ;
}

void ActiveProcessGroup :: erewh(DimensionRep* d) {
  reactivate(d) ;
  onLevel-- ;
}


void ActiveProcessGroup :: where(long _mask) {
  onLevel++ ;

  for(int i = 0 ; i < procs->rnk() ; i++) {
    DimensionRep* d = procs->dim(i).rep ;
    if(_mask & d->mask) {
      if(d->inactive)
        AdlibError("ActiveProcessGroup :: where: "
                   "dimension not in active process array") ;
      else
        deactivate(d) ;
    }
  }
}

void ActiveProcessGroup :: erewh(long _mask) {
  for(int i = 0 ; i < procs->rnk() ; i++) {
    DimensionRep* d = procs->dim(i).rep ;
    if(d->inactive == onLevel)
      reactivate(d) ;
  }

  onLevel-- ;
}


void ActiveProcessGroup :: clip(Group* p) {

  // Replace process group by its intersection with the active process group
  // (support for construction of value sections).

  for(int i = 0 ; i < procs->rnk() ; i++) {
    DimensionRep* d = procs->dim(i).rep ;
    if((p->mask & d->mask) && d->inactive)
      p->deactivate(d, d->coord) ;
  }
}


int ActiveProcessGroup :: size() const {
  // Number of processes in the APG.

  int size = 1 ;
  for(int i = 0 ; i < procs->rnk() ; i++) {
    DimensionRep* dim = procs->dim(i).rep ;
    if(!dim->inactive)
      size *= dim->extent ;
  }

  return size ;
}

int ActiveProcessGroup :: lead() const {

  // Physical id of lead process of APG.

  int parentIndex = 0 ;
  for (int r = 0 ; r < procs->rnk() ; r++) {
    DimensionRep* d = procs->dim(r).rep ;
    if(d->inactive) parentIndex += d->stride * d->crd() ;
  }

  return (procs->physIds) [parentIndex] ;
}


void ActiveProcessGroup :: spanTree(ProcTree* tree) {

  const int index  = locIndex() ;
  const int extent = size() ;

  tree->numChildren = 0 ;
  for(int i = 1 ; i <= PROCTREE_A ; i++) {
    const int ch = PROCTREE_A * index + i ;
    if(ch < extent) {
      tree->child [i - 1] = procId(ch) ;
      tree->numChildren++ ;
    }
  }

  if(index == 0)
    tree->isRoot = 1 ;
  else {
    tree->isRoot = 0 ;
    const int pr = (index - 1) / PROCTREE_A ;
    tree->parent = procId(pr) ;
  }

  tree->root = procId(0) ;
}

int ActiveProcessGroup :: locIndex() const {
  // Index of local process in an enumeration of the APG.

  int index = 0 ;
  for(int i = 0 ; i < procs->rnk() ; i++) {
    DimensionRep* dim = procs->dim(i).rep ;
    if(!dim->inactive)
      index = (dim->extent * index) + dim->crd() ;
  }

  return index ;
}

int ActiveProcessGroup :: procId(const int _index) const {
  // Physical id of process with specified enumeration index in the APG.

  int index = _index ;
  int parentIndex = 0 ;
  for (int i = procs->rnk() - 1 ; i >= 0 ; i--) {
    DimensionRep* dim = procs->dim(i).rep ;
    int coord ;
    if(dim->inactive)
      coord = dim->crd() ;
    else {
      int extent = dim->extent ;
      coord = index % extent ;
      index /= extent ;
    }
    parentIndex += dim->stride * coord ;
  }

  return (procs->physIds) [parentIndex] ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Group'.                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

Group :: Group(const Group& procGrp) : procs(procGrp.procs) {

  // Copy constructor.

  mask = procGrp.mask ;
  leadId = procGrp.leadId ;
}

Group :: Group(const Procs& _procs) : procs((Procs*) &_procs) {
                    // (Procs*) operator casts away `const'


  // Conversion constructor; construct `Group' describing
  // given process array.

  // Check `procs' or `parent' of `procs' is current active process group
  if((procs->parent.procs != apg.procs || procs->parent.mask != apg.mask) &&
     (procs != apg.procs || procs->mask != apg.mask))
    AdlibError("Group :: Group(const Procs&): "
               "inaccessible process array") ;
  
  mask = procs->mask ; ;
  leadId = 0 ;
}

void Group :: deactivate(const DimensionRep* dim, const int coord) {
  
#ifdef CHECK_STRONG_ALIGNMENT
  if(dim->procs != procs)
    AdlibError("Group :: deactivate: "
               "dimension not in process array") ;
#endif

  if(!(mask & dim->mask))
    AdlibError("Group :: deactivate: inactive dimension") ;

  mask   &= ~(dim->mask) ;
  leadId += coord * dim->stride ;
}

void Group :: reactivate(const DimensionRep* dim, const int coord) {
  
  mask   |= dim->mask ;
  leadId -= coord * dim->stride ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Procs': the process array.                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

Procs :: Procs() : rank(1), outsider(0) {            // [rank outsider :-)]

  // Default constructor for `Procs'.  Create the world---a
  // 1-dimensional process array containing the physical processes.

  // Normally this constructor is invoked as part of the `Adlib'
  // startup sequence---not in user code.

  // `parent' left undefined.

  const int coord = AdlibLocId() ;
  const int size  = AdlibNProcs() ;

  physIds = new int [size] ;
  for(int i = 0 ; i < size ; i++)
    physIds [i] = i ;

  dims  = new Dimension ;
  dims->create(size, coord, 1, this, 0) ;

  mask = ~((-1) << rank) ;
}

#include <math.h>

static int root(int n, int d) {

  // integer floor of d'th root of `n'.  n >= 0, d > 0.

  switch(d) {
    case 1 :
      return n ;
    case 2 :
      return (int) sqrt(n) ;
    case 3 :
      return (int) cbrt(n) ;
    default :
      int ceiling  = 1 ;
      int power    = 1 ;
      while(power < n) {
        ceiling++  ;
        power = ceiling ;
        for(int i = 1 ; i < d ; i++)
          power *= ceiling ;
      }
      if(power == n)
        return ceiling ;
      else
        return ceiling - 1 ;
  }
}

static int makeShape(int shp [], int rank, int nprocs) {

  // Compute ``maximal'' shape of specified rank and size <= `nprocs'.

  int size = 1 ;
  while(rank > 0) {
    const int ext = root(nprocs, rank) ;
    rank-- ;
    nprocs /= ext ;
    shp [rank] = ext ;
    size *= ext ;
  }
  return size ;
}

Procs :: Procs(const int _rank, int* _n, int* subIds) :
             rank(_rank), parent(apg) {

  // The standard `Procs' constructor

  // Compute shape and size (`q', `size') of array.

  int* q = new int [rank] ;
  int size ;
  if(_rank == 0 || _n) {
    size = 1 ;
    for (int d = 0; d < rank ; d++) {
      q [d] = _n [d] ;
      size *= _n [d] ;
    }
    if(size > apg.size())
      AdlibError("Procs :: Procs: "
                 "process array too large for active process group\n") ;
  }
  else
    size = makeShape(q, rank, apg.size()) ;

  // Initialize `outsider', `myIndex', and `physIds' vector.

  outsider = 1 ;
  int myIndex = 0 ;
  physIds = new int [size] ;

  const int myApgIndex = apg.locIndex() ;
  for(int i = 0 ; i < size ; i++) {
    int apgIndex = subIds ? subIds [i] : i ;
    if(apgIndex == myApgIndex) {
      outsider = 0 ;
      myIndex  = i ;
    }
    physIds [i] = apg.procId(apgIndex) ;
  }

  // Initialize `dims'.

  dims = new Dimension [rank] ;
  int stride = 1 ;
  for (int d = rank - 1; d >= 0 ; d--) {
    int coord = (myIndex / stride) % q [d] ;
    dims [d].create(q [d], coord, stride, this, d) ;
    stride *= q [d] ;
  }

  delete [] q ;

  mask = ~((-1) << rank) ;
}


Procs :: ~Procs() {
  for(int i = 0 ; i < rank ; i++)
    dims [i].destroy() ;
  delete [] dims ;
  delete [] physIds ;
}

Dimension Procs :: dim(const int i) const {
  if(i < 0 || i >= rank)
    AdlibError("Procs :: dim: axis out of range\n") ;
  return dims [i] ;
}

// these should be inline...

int Procs :: on() {
  return apg.on(this) ;
}

void Procs :: no() {
  apg.no(this) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `DimensionRep': representation class for the process array //
//                          dimension.                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


DimensionRep :: DimensionRep(
	       const int _extent, const int _coord, const int _stride,
	       Procs* _procs, const int _axis) :
    extent(_extent), coord(_coord), stride(_stride),
    procs(_procs),
    axis(_axis),
    inactive(0) {
  mask = 1 << axis ;
}

// ---------------------------------------------------------------------
//
// Things to do...
//
//     Destructor of `Procs' should check if ``on'' process array
//       when it goes out of scope?
//
//     The code for `ActiveProcessGroup' needs revising.  Perhaps take
//       as many checks as possible out to some higher level?
//
//     Make basic functions on `Group' inline.

