
// Schedules for regular remapping of distributed arrays

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"
#include "threads.h"
#include "ProcTree.h"
#include "Procs.h"
#include "Range.h"
#include "SFrame.h"
#include "Frame.h"
#include "schedule.h"
#include "MessSchedule.h"
#include "TreeSchedule.h"
#include "Reduce.h"
#include "Remap.h"


static void shftComponents(int* n, int* exts, int* srcDisps, int* dstDisps,
		           const int extent, const int shift, const Mode mode) ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `RemapSkeleton'                                               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RemapSkeleton :: RemapSkeleton(const Frame* dst, const Frame* src,
                               const int len) :
        BlockMessSchedule(len, src->rnk()), rank(src->rnk()) {
  shapeCheck(dst, src) ;

  exts  = new int [rank] ;
  steps = new int [rank] ;

  Group p = src->grp() ;
  Group q = dst->grp() ;

  // Add send requests to queue

  if(p.on()) {
    DimensionSet rep = p.dims() - src->sig() ;  // Spans replication group.
    rep.where() ;
      if(apg.lead() == p.lead()) {
	rem = dst ;
	loc = src ;
	transfer = &(RemapSkeleton :: send) ;

	loop(0, 1, 1, q, 0) ;
      }
    rep.erewh() ;
  } p.no() ;


  // Add receive requests to queue

  if(q.on()) {
    DimensionSet rep = q.dims() - dst->sig() ;  // Spans replication group.
    rep.where() ;
      if(apg.lead() == q.lead()) {
	rem = src ;
	loc = dst ;
	transfer = &(RemapSkeleton :: recv) ;

	loop(0, 1, 1, p, 0) ;
      }
    rep.erewh() ;
  } q.no() ;

  delete [] steps ;
  delete [] exts ;

  build() ;
}


void RemapSkeleton :: loop(const int offset, const int ctg, const int vol,
                           Group remGrp, const int r) {
  if(r == rank)
    (this->*transfer)(offset, ctg, vol, remGrp.lead()) ;
  else {
    Range  x      = loc->rng(r) ;
    Stride locStr = loc->str(r) ;

    Range  y      = rem->rng(r) ;

    LocBlocksIndex i(x) ;

    // Enumerate locally held blocks of local array range...

    for(i.begin() ; i.test() ; i.next()) {
      //i.sub = i.blockBaseSub ;

      int offsetElem = offset + locStr.offset(i) ;
      const int step = locStr.step(i) ;

      // Local block defines a subrange of local array range.
      // There is a corresponding subrange of the remote array range.
      // Enumerate the blocks in that subrange...

      AllBlocksIndex j(y.subrng(i.blockCount, i.blockBase, i.blockStep)) ;

      for(j.begin() ; j.test() ; j.next()) {
        const int jstep = step * j.blockStep ;

	exts  [r] = j.blockCount ;
        steps [r] = jstep ;

	loop(offsetElem + step * j.blockBase,
             ctg && jstep == vol, vol * j.blockCount, 
             remGrp / j,
             r + 1) ;
      }
    }
  }
}

void RemapSkeleton :: send(int offset, int ctg, int vol, int id) {
  sendReq(offset, steps, ctg, exts, vol, id) ;
}

void RemapSkeleton :: recv(int offset, int ctg, int vol, int id) {
  recvReq(offset, steps, ctg, exts, vol, id) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ShiftSkeleton'                                               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

ShiftSkeleton :: ShiftSkeleton(const Frame* dst, const Frame* src,
                               const int len,
                               const int shift, const int _dim,
                               const Mode mode) :
        BlockMessSchedule(len, src->rnk()),
	dim(_dim), rank(src->rnk()) {

  if(dim < 0 || dim >= src->rnk())
    AdlibError("ShiftSkeleton :: ShiftSkeleton: dim argument out of range.\n") ;
  shapeCheck(dst, src) ;

  Range x = src->rng(dim) ;

  // Set up tables of component contiguous shifts.
  shftComponents(&nShfts, shftExts, shftSrcDisps, shftDstDisps,
		 x.size(), shift, mode) ;

  exts  = new int [rank] ;
  steps = new int [rank] ;

  Group p = src->grp() ;
  if(p.on()) {
    DimensionSet orth = p.dims() - x.dim() ;  // Dims of `p' orthogonal to `x'
    orth.where() ;

      // Add send requests to queue

      loc = src ;
      shftRemDisps = shftDstDisps ;
      shftLocDisps = shftSrcDisps ;
      transfer = &(ShiftSkeleton :: send) ;

      loop(0, 1, 1, apg, 0) ;

      // Add receive requests to queue

      loc = dst ;
      shftRemDisps = shftSrcDisps ;
      shftLocDisps = shftDstDisps ;
      transfer = &(ShiftSkeleton :: recv) ;

      loop(0, 1, 1, apg, 0) ;
    orth.erewh() ;
  } p.no() ;

  delete [] steps ;
  delete [] exts ;

  build() ;
}

void ShiftSkeleton :: loop(const int offset, const int ctg, const int vol,
                           Group remGrp, const int r) {
  if(r == rank)
    (this->*transfer)(offset, ctg, vol, remGrp.lead()) ;
  else {
    Range  x      = loc->rng(r) ;
    Stride locStr = loc->str(r) ;

    if(r == dim)
      for(int i = 0 ; i < nShfts ; i++) {
	const int shftExt     = shftExts     [i] ;
	const int shftLocDisp = shftLocDisps [i] ;
	const int shftRemDisp = shftRemDisps [i] ;

	LocBlocksIndex i(x.subrng(shftExt, shftLocDisp)) ;

	for(i.begin() ; i.test() ; i.next()) {
	  i.sub = i.blockBaseSub ;

	  int offsetElem = offset + locStr.offset(i) ;
	  const int step = locStr.step(i) ;

          AllBlocksIndex j(x.subrng(i.blockCount, shftRemDisp + i.blockBase,
                                    i.blockStep)) ;

	  for(j.begin() ; j.test() ; j.next()) {
            const int jstep = step * j.blockStep ;

	    exts  [r] = j.blockCount ;
	    steps [r] = jstep ;

	    loop(offsetElem + j.blockBase * step,
                 ctg && jstep == vol, vol * j.blockCount,
                 remGrp / j, r + 1) ;
	  }
	}
      }
    else {
      LocBlocksIndex i(x) ;
      for(i.begin() ; i.test() ; i.next()) {
        i.sub = i.blockBaseSub ;

        const int step = locStr.step(i) ;

	steps [r] = step ;
	exts  [r] = i.blockCount ;

	loop(offset + locStr.offset(i), ctg && step == vol, vol * i.blockCount,
             remGrp, r + 1) ;
      }
    }
  }
}

void ShiftSkeleton :: send(int offset, int ctg, int vol, int id) {
  sendReq(offset, steps, ctg, exts, vol, id) ;
}

void ShiftSkeleton :: recv(int offset, int ctg, int vol, int id) {
  recvReq(offset, steps, ctg, exts, vol, id) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Class `SkewSkeleton'                                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

SkewSkeleton :: SkewSkeleton(const Frame* _dst, const Frame* _src,
                             const int len,
                             const Frame* _shf, int* shfDat,
                             const int _dim, const Mode _mode) :
        BlockMessSchedule(len, 1),
	dst(_dst), src(_src), shf(_shf), dim(_dim), mode(_mode) {

  if(dim < 0 || dim >= src->rnk())
    AdlibError("SkewSkeleton :: SkewSkeleton: dim argument out of range.\n") ;
  shapeCheck(dst, src) ;

  rank = src->rnk() ;

  Group p = src->grp() ;
  if(p.on()) {
    Range x = src->rng(dim) ;
    DimensionSet orth = p.dims() - x.dim() ;  // Dims of `p' orthogonal to `x'
    orth.where() ;
      loop(0, 0, shfDat, 0, 0) ;
    orth.erewh() ;
  } p.no() ;

  build() ;
}

void SkewSkeleton :: loop(const int dstOffset, const int srcOffset,
                          int* shfData,
                          const int r, const int s) {

  if(r == rank) {
    Range  x      = src->rng(dim) ;

    Stride dstStr = dst->str(dim) ;
    Stride srcStr = src->str(dim) ;

    int nShfts ;
    int shftExts [2] ;
    int shftSrcDisps [2] ;
    int shftDstDisps [2] ;
    shftComponents(&nShfts, shftExts, shftSrcDisps, shftDstDisps,
		   x.size(), *shfData, mode) ;

    for(int i = 0 ; i < nShfts ; i++) {
      const int shftExt     = shftExts     [i] ;
      const int shftSrcDisp = shftSrcDisps [i] ;
      const int shftDstDisp = shftDstDisps [i] ;

      shift(srcOffset, srcStr, shftExt, shftSrcDisp, shftDstDisp,
            &(SkewSkeleton :: send)) ;
      shift(dstOffset, dstStr, shftExt, shftDstDisp, shftSrcDisp,
            &(SkewSkeleton :: recv)) ;
    }
  }
  else if(r == dim)
    loop(dstOffset, srcOffset, shfData, r + 1, s) ;
  else {
    Range  x      = src->rng(r) ;

    Stride dstStr = dst->str(r) ;
    Stride srcStr = src->str(r) ;

    Stride shfStr = shf->str(s) ;

    LocBlocksIndex i(x) ;

    for(i.begin() ; i.test() ; i.next()) {
      i.sub = i.blockBaseSub ;

      int dstOffsetElem = dstOffset + dstStr.offset(i) ;
      int dstStep = dstStr.step(i) ;

      int srcOffsetElem = srcOffset + srcStr.offset(i) ;
      int srcStep = srcStr.step(i) ;

      int* shfDataElem = shfData + shfStr.offset(i) ;
      int shfStep = shfStr.step(i) ;

      for(int sub = 0 ; sub < i.blockCount ; sub++) {
	loop(dstOffsetElem, srcOffsetElem, shfDataElem, r + 1, s + 1) ;

        dstOffsetElem += dstStep ;
        srcOffsetElem += srcStep ;
        shfDataElem += shfStep ;
      }
    }
  }
}

void SkewSkeleton :: shift(int offset, Stride str,
                           const int ext, const int locDisp, const int remDisp,
                       void (SkewSkeleton :: *transfer)(int, int, int, int)) {

  Range x = src->rng(dim) ;

  LocBlocksIndex i(x.subrng(ext, locDisp)) ;
  for(i.begin() ; i.test() ; i.next()) {
    i.sub = i.blockBaseSub ;

    int offsetElem = offset + str.offset(i) ;
    int step = str.step(i) ;

    AllBlocksIndex j(x.subrng(i.blockCount, remDisp + i.blockBase,
                              i.blockStep)) ;

    for(j.begin() ; j.test() ; j.next()) {
      Group p = apg / j ; 
      (this->*transfer)(offsetElem + j.blockBase * step, step * j.blockStep,
                        j.blockCount, p.lead()) ;
    }
  }
}

void SkewSkeleton :: send(int offset, int step, int count, int id) {
  sendReq(offset, &step, step == 1, &count, count, id) ;
}

void SkewSkeleton :: recv(int offset, int step, int count, int id) {
  recvReq(offset, &step, step == 1, &count, count, id) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `MultiShiftSkeleton'                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

MultiShiftSkeleton :: MultiShiftSkeleton(const Frame* dst, const Frame* src,
                                         const int len,
                                         int* shift, Mode* mode) :
        BlockMessSchedule(len, src->rnk()),
	rank(src->rnk()) {

  shapeCheck(dst, src) ;

  nShfts       = new int [rank] ;
  shftExts     = new int [rank] [2] ;
  shftSrcDisps = new int [rank] [2] ;
  shftDstDisps = new int [rank] [2] ;

  // Set up tables of component contiguous shifts.
  for(int r = 0 ; r < rank ; r++)
    shftComponents(&(nShfts [r]),
                   shftExts [r], shftSrcDisps [r], shftDstDisps [r],
		   src->rng(r).size(), shift [r], mode [r]) ;

  exts  = new int [rank] ;
  steps = new int [rank] ;

  Group p = src->grp() ;
  if(p.on()) {
    DimensionSet rep = p.dims() - src->sig() ;  // Spans replication group.
    rep.where() ;
      // Add send requests to queue

      loc = src ;
      shftRemDisps = shftDstDisps ;
      shftLocDisps = shftSrcDisps ;
      transfer = &(MultiShiftSkeleton :: send) ;

      loop(0, 1, 1, apg, 0) ;

      // Add receive requests to queue

      loc = dst ;
      shftRemDisps = shftSrcDisps ;
      shftLocDisps = shftDstDisps ;
      transfer = &(MultiShiftSkeleton :: recv) ;

      loop(0, 1, 1, apg, 0) ;
    rep.erewh() ;
  } p.no() ;

  delete [] steps ;
  delete [] exts ;

  delete [] nShfts ;
  delete [] shftExts ;
  delete [] shftSrcDisps ;
  delete [] shftDstDisps ;

  build() ;
}

void MultiShiftSkeleton :: loop(const int offset, const int ctg, const int vol,
                                Group remGrp, const int r) {
  if(r == rank)
    (this->*transfer)(offset, ctg, vol, remGrp.lead()) ;
  else {
    Range  x      = loc->rng(r) ;
    Stride locStr = loc->str(r) ;

    for(int shf = 0 ; shf < nShfts [r] ; shf++) {
      const int shftExt     = shftExts     [r] [shf] ;
      const int shftLocDisp = shftLocDisps [r] [shf] ;
      const int shftRemDisp = shftRemDisps [r] [shf] ;

      LocBlocksIndex i(x.subrng(shftExt, shftLocDisp)) ;

      for(i.begin() ; i.test() ; i.next()) {
	i.sub = i.blockBaseSub ;

	int offsetElem = offset + locStr.offset(i) ;
	const int step = locStr.step(i) ;

	steps [r] = step ;

        AllBlocksIndex j(x.subrng(i.blockCount, shftRemDisp + i.blockBase)) ;

	for(j.begin() ; j.test() ; j.next()) {
	  const int count = j.blockCount ;

	  exts [r] = count ;

	  loop(offsetElem, ctg && step == vol, vol * count,
               remGrp / j, r + 1) ;

	  offsetElem += count * step ;
	}
      }
    }
  }
}

void MultiShiftSkeleton :: send(int offset, int ctg, int vol, int id) {
  sendReq(offset, steps, ctg, exts, vol, id) ;
}

void MultiShiftSkeleton :: recv(int offset, int ctg, int vol, int id) {
  recvReq(offset, steps, ctg, exts, vol, id) ;
}


static void shftComponents(int* n, int* exts, int* srcDisps, int* dstDisps,
		           const int extent, const int shift, const Mode mode) {
  *n = 0 ;

  switch(mode) {
    case CYCL : {
      const int shiftMod =
	  shift < 0 ? (shift % extent + extent) : (shift % extent) ;

      exts     [*n] = extent - shiftMod ;
      srcDisps [*n] = shiftMod ;
      dstDisps [*n] = 0 ;
      (*n)++ ;

      if(shiftMod != 0) {
	exts     [*n] = shiftMod ;
	srcDisps [*n] = 0 ;
	dstDisps [*n] = extent - shiftMod ;
	(*n)++ ;
      }
                  

      break ;
    }
    case EDGE : {
      if(shift >= 0 && shift < extent) {
	exts     [*n] = extent - shift ;
	srcDisps [*n] = shift ;
	dstDisps [*n] = 0 ;
	(*n)++ ;
      }
      
      if(shift < 0 && shift > -extent) {
	exts     [*n] = extent + shift ;
	srcDisps [*n] = 0 ;
	dstDisps [*n] = -shift ;
	(*n)++ ;
      }

      break ;
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `WriteFaces'.  Write upper and lower ghost areas for          //
//                      dimension `dim'.                               //
//                                                                     //
//                      Copy includes ghost extensions associated with //
//                      dimensions r < dim.  (But not those associated //
//                      with dimensions r > dim.)                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }

WriteFaces :: WriteFaces(const Frame* _src, const int len, const int _dim,
                         const int* _wlo, const int* _whi) :
        BlockMessSchedule(len, _src->rnk()),
        src(_src), dim(_dim), wlo(_wlo), whi(_whi), rank(_src->rnk()) {

  Range x = src->rng(dim) ;
  Group p = src->grp() ;

  exts  = new int [rank] ;
  steps = new int [rank] ;

  if(p.on()) {
    DimensionSet orth = p.dims() - x.dim() ;  // Dims of `p' orthogonal to `x'
    orth.where() ;
      loop(0, 1, 1, apg, 0) ;
    orth.erewh() ;
  } p.no() ;

  delete [] steps ;
  delete [] exts ;

  build() ;
}

void WriteFaces :: loop(const int offset,
                        const int ctg, const int vol,
		        const Group& remGrp, const int r) {

  Range  x      = src->rng(r) ;
  Stride locStr = src->str(r) ;

  LocBlocksIndex i(x) ;
  if(r == dim) {
    for(i.begin() ; i.test() ; i.next()) {
      int offsetElem = offset + locStr.offset(i) ;
      const int step = locStr.step(i) ;

      if(wlo [r] > 0) {

        // Sending data to lower ghost edges.

	// Iterate over a sub-range of width `wlor' to the right of the
	// current block.  This includes start points of all blocks whose
        // lower ghost area is covered by the current block.

	const int neighbBase = i.blockBase + i.blockCount ;
	const int wlor = min(wlo [r], x.size() - neighbBase) ;
	AllBlocksIndex jlor(x.subrng(wlor, neighbBase)) ;

	for(jlor.begin() ; jlor.test() ; jlor.next()) {

	  // Lower ghost region for (parent block of) `jlor' is
	  //     x.subrng(wlo [r], neighbBase + jlor.blockBase - wlo [r])
	  // Send overlap of current block (`i') and this ghost region
	  // to owner of block `jlor'.

	  const int start = max(i.blockCount + jlor.blockBase - wlo [r], 0) ;
	  const int count = i.blockCount - start ;

	  exts  [r] = count ;
	  steps [r] = step ;

	  // offset + locStr.offset(i >> start)
	  sendLoop(offsetElem + start * step,
		   ctg && step == vol, vol * count,
		   remGrp / jlor, r + 1) ;
	}

	// Receiving data for lower ghost edges.

	// Iterate over a sub-range of width `wlol' to the left of the
	// current block.

	const int wlol = min(wlo [r], i.blockBase) ;
	AllBlocksIndex jlol(x.subrng(wlol, i.blockBase - wlol)) ;

	for(jlol.begin() ; jlol.test() ; jlol.next()) {
	  exts  [r] = jlol.blockCount ;
	  steps [r] = step ;

	  // offset + locStr.offset(i << (wlol - jlol.blockBase))
	  recvLoop(offsetElem + (jlol.blockBase - wlol) * step,
		   ctg && step == vol, vol * jlol.blockCount,
		   remGrp / jlol, r + 1) ;
	}
      }

      if(whi [r] > 0) {

	// Sending data to upper ghost edges.

	// Iterate over a sub-range of width `whil' to the left of the
	// current block.  This includes end points of all blocks whose
        // upper ghost area is covered by the current block.

	const int whil = min(whi [r], i.blockBase) ;
	AllBlocksIndex jhil(x.subrng(whil, i.blockBase - whil)) ;

	for(jhil.begin() ; jhil.test() ; jhil.next()) {

	  // Upper ghost region for (parent block of) `jhil' is
	  //     x.subrng(whi [r], i.blockBase - whil +
	  //                       jhil.blockBase + jhil.blockCount)
	  // Send overlap of current block (`i') and this ghost region
	  // to owner of block `jhil'.

	  const int count = min(whi [r] -
                                    (whil - jhil.blockBase - jhil.blockCount),
		                i.blockCount) ;

	  exts  [r] = count ;
	  steps [r] = step ;

	  sendLoop(offsetElem,
		   ctg && step == vol, vol * count,
		   remGrp / jhil, r + 1) ;
	}

	// Receiving data for upper ghost edges.

	// Iterate over a sub-range of width `whir' to the right of the
	// current block.

	const int neighbBase = i.blockBase + i.blockCount ;
	const int whir = min(whi [r], x.size() - neighbBase) ;
	AllBlocksIndex jhir(x.subrng(whir, neighbBase)) ;

	for(jhir.begin() ; jhir.test() ; jhir.next()) {
	  exts  [r] = jhir.blockCount ;
	  steps [r] = step ;

	  // offset + locStr.offset(i >> (i.blockCount + jhir.blockBase))
	  recvLoop(offsetElem + (i.blockCount + jhir.blockBase) * step,
		   ctg && step == vol, vol * jhir.blockCount,
		   remGrp / jhir, r + 1) ;
	}
      }
    }
  }
  else {

    // r < dim.  Include ghost extensions for orthogonal dimension in copy.

    for(i.begin() ; i.test() ; i.next()) {
      const int step  = locStr.step(i) ;
      const int count = i.blockCount + wlo [r] + whi [r] ;

      steps [r] = step ;
      exts  [r] = count ;

      // offset + locStr.offset(i << wlo [r])
      loop(offset + locStr.offset(i) - wlo [r] * step,
	   ctg && step == vol, vol * count,
	   remGrp, r + 1) ;
    }
  }
}

void WriteFaces :: recvLoop(const int offset,
                            const int ctg, const int vol,
                            const Group& remGrp, const int r) {
  if(r == rank)
    recvReq(offset, steps, ctg, exts, vol, remGrp.lead()) ;
  else {
    Range  x      = src->rng(r) ;
    Stride locStr = src->str(r) ;

    LocBlocksIndex i(x) ;
    for(i.begin() ; i.test() ; i.next()) {
      const int step = locStr.step(i) ;

      steps [r] = step ;
      exts  [r] = i.blockCount ;

      recvLoop(offset + locStr.offset(i),
	       ctg && step == vol, vol * i.blockCount,
	       remGrp, r + 1) ;
    }
  }
}

void WriteFaces :: sendLoop(const int offset,
                            const int ctg, const int vol,
                            const Group& remGrp, const int r) {
  if(r == rank)
    sendReq(offset, steps, ctg, exts, vol, remGrp.lead()) ;
  else {
    Range  x      = src->rng(r) ;
    Stride locStr = src->str(r) ;

    LocBlocksIndex i(x) ;
    for(i.begin() ; i.test() ; i.next()) {
      const int step = locStr.step(i) ;

      steps [r] = step ;
      exts  [r] = i.blockCount ;

      sendLoop(offset + locStr.offset(i),
	       ctg && step == vol, vol * i.blockCount,
	       remGrp, r + 1) ;
    }
  }
}

// -------------------------------------------------------------
// Things to do
//
//     Where there are several short `int' vectors in a schedule (`exts',
//     `steps', ...) allocate and deallocate them as a single vector.
//
//     Or allocate them using `Vec'?

