
// Operations on ``step-wise'' block distributed ranges (C* support)

#include <stdlib.h>
#include <stdio.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepRangeRep': stepped block-distributed range (C* format).  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

StepRangeRep :: StepRangeRep(int _extent, Dimension _dim,
                             const Layout _layout) :
                    RangeRep(_extent, _dim), layout(_layout) {
  blockSize = extent / dim.size() ;
  stepCoord = extent % dim.size() ;
}

StepRangeRep :: StepRangeRep(int _extent, Dimension _dim,
                             const int _blockSize, const int _stepCoord,
                             const Layout _layout) :
                    RangeRep(_extent, _dim),
                    blockSize(_blockSize), stepCoord(_stepCoord),
                    layout(_layout) {}
 

RangeRep* StepRangeRep :: prm() {return this ; }
int StepRangeRep :: bas() {return 0 ; }
int StepRangeRep :: str() {return 1 ; }

//int StepRangeRep :: isScalar() {return 0 ; }


inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }


// Subscripting of Range.

Subscript* StepRangeRep :: subscr(const int _i) {
  if(_i < stepCoord * (blockSize + 1)) {
    crd = _i / (blockSize + 1) ;
    sub = _i - (blockSize + 1) * crd ;
  }
  else {
    const int redLoc = _i - stepCoord ;

    crd = redLoc / blockSize ;
    sub = redLoc - blockSize * crd ;
  }

  return this ;
}

int StepRangeRep :: subscr(const Subscript* s) {
  if(crd < stepCoord)
    return (blockSize + 1) * s->crd + s->sub ;
  else
    return blockSize * s->crd + stepCoord + s->sub ;
}


// Creation of child ranges and strides...

RangeRep* StepRangeRep :: subrng(const int _extent,
                                 const int _base, const int _stride) {
  return new StepSubRangeRep(_extent, _base, _stride, this,
                             dim, blockSize, stepCoord, layout) ;
}

StrideRep* StepRangeRep :: vector(const int dataLen) {
  if(stepCoord == 0)
    return new SimpleStrideRep(dataLen, blockSize) ;
  else
    return new SimpleStrideRep(dataLen, blockSize + 1) ;
}


// Auxilliary functions for use during block iteration

void StepRangeRep :: blockInit(Block& b) {
  b.dim          = dim ;

  b.blockStep    = 1 ;

  b.blockStepSub = 1 ;
}

void StepRangeRep :: blockParams(Block& b, int i_lo, int i_hi) {
  b.blockCount   = i_hi - i_lo + 1 ;

  b.blockBase    = i_lo ;

  b.blockBaseSub = 0 ;
}


// Iteration over locally held blocks.

void StepRangeRep :: locBlockInit(Block& b) {
  int i_lo, i_hi ;

  b.crd          = dim.crd() ;

  if(b.crd < stepCoord) {
    i_lo = b.crd * (blockSize + 1) ;
    i_hi = i_lo + blockSize ;
  }
  else {
    i_lo = b.crd * blockSize + stepCoord ;
    i_hi = i_lo + blockSize - 1 ;
  }

  if(i_lo > i_hi)
    b.blockCount = 0 ;
  else
    blockParams(b, i_lo, i_hi) ;
}


// Iteration over all blocks of range.

void StepRangeRep :: allBlocksEnum(Block& b) {
  int i_lo, i_hi ;

  b.crd = 0 ;

  i_lo    = 0 ;

  if(b.crd < stepCoord)
    i_hi = blockSize ;
  else
    i_hi = blockSize - 1 ;

  blockParams(b, i_lo, i_hi) ;
}

int StepRangeRep :: allBlocksTest(Block& b) {return b.blockBase < extent ; }

void StepRangeRep :: allBlocksNext(Block& b) {
  int i_lo, i_hi ;

  b.crd++ ;

  if(b.crd < stepCoord) {
    i_lo = b.blockBase + blockSize + 1 ;
    i_hi = i_lo + blockSize ;
  }
  else {
    i_lo = b.blockBase + blockSize ;
    i_hi = i_hi + blockSize - 1 ;
  }

  blockParams(b, i_lo, i_hi) ;
}


// Memory management

MEMORY_POOL_DEFN(StepRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepSubRangeRep': range aligned to a stepped                 //
//                          block-distributed range.                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

StepSubRangeRep :: StepSubRangeRep(const int _extent,
                       const int _base, const int _stride,
                       StepRangeRep* _parent,
                       Dimension _dim, const int _blockSize,
                       const int _stepCoord, const Layout _layout) :
                           StepRangeRep(_extent, _dim,
                                        _blockSize, _stepCoord, _layout),
                           base(_base), stride(_stride),
                           parent(_parent) {
  parent->incr() ;
}

StepSubRangeRep :: ~StepSubRangeRep() {
  if(parent->decr() == 0) delete parent ;
}

RangeRep* StepSubRangeRep :: prm() {return parent ; }
int StepSubRangeRep :: bas() {return base ; }
int StepSubRangeRep :: str() {return stride ; }


// Subscripting of Range.

Subscript* StepSubRangeRep :: subscr(const int _i) {
  return StepRangeRep :: subscr(base + stride * _i) ;
}

int StepSubRangeRep :: subscr(const Subscript* s) {
  return (StepRangeRep :: subscr(s) - base) / stride ;
}


// Creation of child ranges and strides...

RangeRep* StepSubRangeRep :: subrng(const int _extent,
                                    const int _base, const int _stride) {
  return new StepSubRangeRep(_extent,
                             base + stride * _base, stride * _stride, parent,
                             dim, blockSize, stepCoord, layout) ;
}

StrideRep* StepSubRangeRep :: vector(const int dataLen) {
  if(layout == STANDARD) 
    return StepRangeRep :: vector(dataLen) ; 
  else {

    // layout == PACKED 

    if(stride == 0) AdlibError("StepSubRangeRep :: vector: "
			       "range stride is zero") ;

    int blockWidth, memStep ;

    if(stepCoord == 0)
      blockWidth = blockSize ;
    else
      blockWidth = blockSize + 1 ;

    if(stride > 0) {
      blockWidth = (blockWidth + stride - 1) / stride ;

      memStep = dataLen ;
    }
    else {
      blockWidth = (blockWidth - stride - 1) / (-stride) ;

      memStep = -dataLen ;
    }

    return new PackedStrideRep(memStep, stride, blockWidth) ;
  }
}


// Auxilliary functions for use during block iteration

void StepSubRangeRep :: blockInit(Block& b) {
  b.dim          = dim ;

  b.blockStep    = 1 ;

  b.blockStepSub = stride ;
}

void StepSubRangeRep :: blockParams(Block& b, int i_lo, int i_hi) {
  b.blockCount   = i_hi - i_lo + 1 ;

  b.blockBase    = i_lo ;

  if(b.crd < stepCoord)
    b.blockBaseSub = base + stride * i_lo - (blockSize + 1) * b.crd ;
  else
    b.blockBaseSub = base + stride * i_lo - stepCoord - blockSize * b.crd ;
}

void StepSubRangeRep :: blockRange(int& i_lo, int& i_hi, int _crd) {

  // Define `i_lo' and `i_hi' for current value of `_crd'

  // If `_crd' does not define a position within the bounds of
  // the range, the results are undefined but it is guaranteed that
  // i_hi < i_lo

  int blockStartOffset  ;
  int blockFinishOffset ;

  if(_crd < stepCoord) {
    const int blockStart = _crd * (blockSize + 1) ;

    blockStartOffset  = blockStart - base ;
    blockFinishOffset = blockStartOffset + blockSize ;
  }
  else {
    const int blockStart = _crd * blockSize + stepCoord ;

    blockStartOffset  = blockStart - base ;
    blockFinishOffset = blockStartOffset + blockSize - 1 ;
  }

  if(stride > 0) {
    i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
    if(blockFinishOffset >= 0)
      i_hi = min(blockFinishOffset / stride, extent - 1) ;
    else
      i_hi = -1 ;
  }
  else {
    i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
    if(blockStartOffset <= 0)
      i_hi = min(blockStartOffset / stride, extent - 1) ;
    else
      i_hi = -1 ;
  }
}


// Iteration over locally held blocks.

void StepSubRangeRep :: locBlockInit(Block& b) {
  int i_lo, i_hi ;

  b.crd          = dim.crd() ;

  blockRange(i_lo, i_hi, b.crd) ;

  if(i_lo > i_hi)
    b.blockCount = 0 ;
  else
    blockParams(b, i_lo, i_hi) ;
}


// Iteration over all blocks of range.

void StepSubRangeRep :: allBlocksEnum(Block& b) {
  int i_lo, i_hi ;

  if(base < stepCoord * (blockSize + 1))
    b.crd = base / (blockSize + 1) ;
  else
    b.crd = (base - stepCoord) / blockSize ;

  blockRange(i_lo, i_hi, b.crd) ;

  if(i_lo > i_hi && i_lo < extent)
    allBlocksNext(b) ;
  else
    blockParams(b, i_lo, i_hi) ;
}

void StepSubRangeRep :: allBlocksNext(Block& b) {
  int i_lo, i_hi ;

  do {
    if(stride > 0)
      b.crd++ ;
    else
      b.crd-- ;

    blockRange(i_lo, i_hi, b.crd) ;

  } while (i_lo > i_hi && i_lo < extent) ;

  blockParams(b, i_lo, i_hi) ;
}


// Memory management

MEMORY_POOL_DEFN(StepSubRangeRep)

